from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
from numpy.typing import NDArray

from ._geneticalgorithm import GeneticAlgorithm
from ..utils import numpy_group_by
from ..utils.random import random_weighted_sample


class SelfCGA(GeneticAlgorithm):
    """Semenkin, E.S., Semenkina, M.E. Self-configuring Genetic Algorithm with Modified Uniform
    Crossover Operator. LNCS, 7331, 2012, pp. 414-421. https://doi.org/10.1007/978-3-642-30976-2_50
    """

    def __init__(
        self,
        fitness_function: Callable[[NDArray[Any]], NDArray[np.float64]],
        iters: int,
        pop_size: int,
        str_len: int,
        tour_size: int = 2,
        mutation_rate: float = 0.05,
        parents_num: int = 2,
        elitism: bool = True,
        selections: Tuple[str, ...] = (
            "proportional",
            "rank",
            "tournament_3",
            "tournament_5",
            "tournament_7",
        ),
        crossovers: Tuple[str, ...] = (
            "empty",
            "one_point",
            "two_point",
            "uniform_2",
            "uniform_7",
            "uniform_prop_2",
            "uniform_prop_7",
            "uniform_rank_2",
            "uniform_rank_7",
            "uniform_tour_3",
            "uniform_tour_7",
        ),
        mutations: Tuple[str, ...] = ("weak", "average", "strong"),
        init_population: Optional[NDArray[np.byte]] = None,
        K: float = 2,
        selection_threshold_proba: float = 0.05,
        crossover_threshold_proba: float = 0.05,
        mutation_threshold_proba: float = 0.05,
        genotype_to_phenotype: Optional[Callable[[NDArray[np.byte]], NDArray[Any]]] = None,
        optimal_value: Optional[float] = None,
        termination_error_value: float = 0.0,
        no_increase_num: Optional[int] = None,
        minimization: bool = False,
        show_progress_each: Optional[int] = None,
        keep_history: bool = False,
        n_jobs: int = 1,
        fitness_function_args: Optional[Dict] = None,
        genotype_to_phenotype_args: Optional[Dict] = None,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        on_generation: Optional[Callable] = None,
    ):
        GeneticAlgorithm.__init__(
            self,
            fitness_function=fitness_function,
            iters=iters,
            pop_size=pop_size,
            str_len=str_len,
            tour_size=tour_size,
            mutation_rate=mutation_rate,
            parents_num=parents_num,
            elitism=elitism,
            init_population=init_population,
            genotype_to_phenotype=genotype_to_phenotype,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history,
            n_jobs=n_jobs,
            fitness_function_args=fitness_function_args,
            genotype_to_phenotype_args=genotype_to_phenotype_args,
            random_state=random_state,
            on_generation=on_generation,
        )

        self._K: float = K
        self._thresholds: Dict[str, float] = {
            "selection": selection_threshold_proba,
            "crossover": crossover_threshold_proba,
            "mutation": mutation_threshold_proba,
        }

        self._selection_set: Dict[str, Tuple[Callable, Union[float, int]]] = {}
        self._crossover_set: Dict[str, Tuple[Callable, Union[float, int]]] = {}
        self._mutation_set: Dict[str, Tuple[Callable, Union[float, int], bool]] = {}

        self._selection_proba: Dict[str, float]
        self._crossover_proba: Dict[str, float]
        self._mutation_proba: Dict[str, float]

        for operator_name in selections:
            self._selection_set[operator_name] = self._selection_pool[operator_name]
        self._selection_set = dict(sorted(self._selection_set.items()))

        for operator_name in crossovers:
            self._crossover_set[operator_name] = self._crossover_pool[operator_name]
        self._crossover_set = dict(sorted(self._crossover_set.items()))

        for operator_name in mutations:
            self._mutation_set[operator_name] = self._mutation_pool[operator_name]
        self._mutation_set = dict(sorted(self._mutation_set.items()))

        self._z_selection = len(self._selection_set)
        self._z_crossover = len(self._crossover_set)
        self._z_mutation = len(self._mutation_set)

        self._selection_proba = dict(
            zip(list(self._selection_set.keys()), np.full(self._z_selection, 1 / self._z_selection))
        )
        if "empty" in self._crossover_set.keys():
            self._crossover_proba = dict(
                zip(
                    list(self._crossover_set.keys()),
                    np.full(self._z_crossover, 0.9 / (self._z_crossover - 1)),
                )
            )
            self._crossover_proba["empty"] = 0.1
        else:
            self._crossover_proba = dict(
                zip(
                    list(self._crossover_set.keys()),
                    np.full(self._z_crossover, 1 / self._z_crossover),
                )
            )
        self._mutation_proba = dict(
            zip(list(self._mutation_set.keys()), np.full(self._z_mutation, 1 / self._z_mutation))
        )
        self._selection_operators: NDArray
        self._crossover_operators: NDArray
        self._mutation_operators: NDArray

    def _choice_operators(self: SelfCGA, proba_dict: Dict["str", float]) -> NDArray:
        operators = np.array(list(proba_dict.keys()))
        proba = np.array(list(proba_dict.values()), dtype=np.float64)

        index = random_weighted_sample(weights=proba, quantity=self._pop_size, replace=True)
        chosen_operator = operators[index]
        return chosen_operator

    def _get_new_proba(
        self: SelfCGA,
        proba_dict: Dict["str", float],
        operator: str,
        threshold: float,
    ) -> Dict["str", float]:
        proba_dict[operator] += self._K / self._iters
        proba_value = np.array(list(proba_dict.values()))
        proba_value -= self._K / (len(proba_dict) * self._iters)
        proba_value = proba_value.clip(threshold, 1)
        proba_value = proba_value / proba_value.sum()
        new_proba_dict = dict(zip(proba_dict.keys(), proba_value))
        return new_proba_dict

    def _find_fittest_operator(
        self: SelfCGA, operators: NDArray, fitness: NDArray[np.float64]
    ) -> str:
        keys, groups = numpy_group_by(group=fitness, by=operators)
        mean_fit = np.array(list(map(np.mean, groups)))
        fittest_operator = keys[np.argmax(mean_fit)]
        return fittest_operator

    def _update_data(self: SelfCGA) -> None:
        super()._update_data()
        self._update_stats(
            s_proba=self._selection_proba,
            c_proba=self._crossover_proba,
            m_proba=self._mutation_proba,
        )

    def _adapt(self: SelfCGA) -> None:
        s_fittest_oper = self._find_fittest_operator(self._selection_operators, self._fitness_i)
        self._selection_proba = self._get_new_proba(
            self._selection_proba, s_fittest_oper, self._thresholds["selection"]
        )

        c_fittest_oper = self._find_fittest_operator(self._crossover_operators, self._fitness_i)
        self._crossover_proba = self._get_new_proba(
            self._crossover_proba, c_fittest_oper, self._thresholds["crossover"]
        )

        m_fittest_oper = self._find_fittest_operator(self._mutation_operators, self._fitness_i)
        self._mutation_proba = self._get_new_proba(
            self._mutation_proba, m_fittest_oper, self._thresholds["mutation"]
        )

        self._selection_operators = self._choice_operators(proba_dict=self._selection_proba)
        self._crossover_operators = self._choice_operators(proba_dict=self._crossover_proba)
        self._mutation_operators = self._choice_operators(proba_dict=self._mutation_proba)

    def _get_new_population(self: SelfCGA) -> None:
        self._population_g_i = np.array(
            [
                self._get_new_individ_g(
                    self._selection_operators[i],
                    self._crossover_operators[i],
                    self._mutation_operators[i],
                )
                for i in range(self._pop_size)
            ],
            dtype=self._population_g_i.dtype,
        )

    def _get_init_population(self: SelfCGA) -> None:
        self._selection_operators = self._choice_operators(proba_dict=self._selection_proba)
        self._crossover_operators = self._choice_operators(proba_dict=self._crossover_proba)
        self._mutation_operators = self._choice_operators(proba_dict=self._mutation_proba)
        return super()._get_init_population()



from __future__ import annotations

from collections import Counter
from collections import defaultdict
from inspect import signature
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from operator import add
from operator import sub
from operator import mul
from operator import abs
from operator import gt

import hashlib

import numpy as np
from numpy.typing import NDArray

from ._net import HiddenBlock
from ..utils import find_end_subtree_from_i
from ..utils import find_id_args_from_i
from ..utils import get_levels_tree_from_i
from ..utils import common_region
from ..utils import common_region_two_trees
from ..utils import save_div
from ..utils import save_exp
from ..utils import sqrtabs
from ..utils import logabs
from ..utils import sin
from ..utils import cos
from ..utils import create_operator
from ..utils.random import random_sample
from ..utils.random import random_weighted_sample
from ..utils.random import uniform


FUNCTIONAL_COLOR_CODE = (1, 0.72, 0.43, 1)
TERMINAL_COLOR_CODE = (0.21, 0.76, 0.56, 1)


class Node:
    def __init__(self, value: Any, name: str, sign: str, n_args: int) -> None:
        self._value = value
        self._name = name
        self._sign = sign
        self._n_args = n_args

    def __str__(self) -> str:
        return str(self._sign)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Node):
            return NotImplemented
        return self._name == other._name


class FunctionalNode(Node):
    def __init__(self, value: Any, sign: Optional[str] = None) -> None:
        Node.__init__(
            self,
            value=value,
            name=value.__name__,
            sign=sign or value._sign,
            n_args=len(signature(value._operation).parameters),
        )


class TerminalNode(Node):
    def __init__(self, value: Any, name: str) -> None:
        Node.__init__(self, value=value, name=name, sign=name, n_args=0)


class EphemeralConstantNode(TerminalNode):
    def __init__(self, value: Any, name: str) -> None:
        TerminalNode.__init__(self, value=value, name=name)


class EphemeralNode(Node):
    def __init__(self, generator: Callable) -> None:
        self._generator = generator
        Node.__init__(
            self,
            value=self._generator,
            name=str(self._generator.__name__),
            sign=str(self._generator.__name__),
            n_args=0,
        )

    def __call__(self) -> EphemeralConstantNode:
        value = self._generator()
        node = EphemeralConstantNode(value=value, name=str(value))
        return node


class UniversalSet:
    def __init__(
        self,
        functional_set: Tuple[FunctionalNode, ...],
        terminal_set: Tuple[Union[TerminalNode, EphemeralNode], ...],
    ) -> None:
        self._functional_set = self._define_functional_set(functional_set)
        self._terminal_set = tuple(terminal_set)

    def _define_functional_set(self, functional_set: Tuple[FunctionalNode, ...]) -> defaultdict:
        _functional_set_list = defaultdict(list, {-1: list(functional_set)})
        _functional_set_tuple = defaultdict(tuple)
        for unit in functional_set:
            _functional_set_list[unit._n_args].append(unit)
        for key, value in _functional_set_list.items():
            _functional_set_tuple[key] = tuple(value)
        return _functional_set_tuple

    def _random_terminal_or_ephemeral(self) -> Union[EphemeralConstantNode, TerminalNode]:
        index = random_sample(len(self._terminal_set), 1, True)[0]
        choosen = self._terminal_set[index]
        if isinstance(choosen, EphemeralNode):
            return choosen()
        else:
            return choosen

    def _random_functional(self, n_args: int = -1) -> FunctionalNode:
        n_args_functionals = self._functional_set[n_args]
        index = random_sample(len(n_args_functionals), 1, True)[0]
        choosen = n_args_functionals[index]
        return choosen


class EnsembleUniversalSet(UniversalSet):
    def __init__(
        self,
        functional_set: Tuple[FunctionalNode, ...],
        terminal_set: Tuple[Union[TerminalNode, EphemeralNode], ...],
    ) -> None:
        UniversalSet.__init__(self, functional_set, terminal_set)
        self._functional_set_proba = self._define_functional_set_proba()

    def _define_functional_set_proba(self: EnsembleUniversalSet) -> Dict[int, NDArray[np.float64]]:
        functional_set_proba: Dict[int, NDArray[np.float64]] = defaultdict(
            lambda: np.array([], dtype=np.float64)
        )
        for n_args, value in self._functional_set.items():
            count = Counter([type(node._value) for node in value])
            for node in value:
                type_ = type(node._value)
                functional_set_proba[n_args] = np.append(functional_set_proba[n_args], count[type_])
            proba = functional_set_proba[n_args]
            functional_set_proba[n_args] = 1 / proba
        return functional_set_proba

    def _random_functional(self, n_args: int = -1) -> FunctionalNode:
        n_args_functionals = self._functional_set[n_args]
        weights = self._functional_set_proba[n_args]
        index = random_weighted_sample(weights=weights, quantity=1, replace=True)[0]
        choosen = n_args_functionals[index]

        return choosen


class Tree:
    def __init__(
        self,
        nodes: List[Union[FunctionalNode, TerminalNode, EphemeralNode]],
        n_args: Optional[Union[List[int], NDArray[np.int64]]] = None,
    ) -> None:
        self._nodes = nodes
        if n_args is None:
            self._n_args = self._init_n_args()
        else:
            self._n_args = np.array(n_args, dtype=np.int64)

    def _init_n_args(self) -> NDArray[np.int64]:
        n_args = np.empty(shape=len(self._nodes), dtype=np.int64)
        for i, node_i in enumerate(self._nodes):
            n_args[i] = node_i._n_args
        return n_args

    def __len__(self) -> int:
        return len(self._nodes)

    def __str__(self) -> str:
        pack: List[str] = []
        for node in reversed(self._nodes):
            args = []
            for _ in range(node._n_args):
                args.append(pack.pop())
            if isinstance(node, FunctionalNode):
                pack.append(node._value._write(*args))
            else:
                pack.append(node._name)
        return pack[0]

    def signature(self) -> str:
        s = str(self)
        return hashlib.sha1(s.encode()).hexdigest()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Tree):
            raise TypeError(f"Cannot compare Tree with {type(other).__name__}")
        if len(self) != len(other):
            return False
        for node_1, node_2 in zip(self._nodes, other._nodes):
            if node_1 != node_2:
                return False
        return True

    def __call__(self) -> Any:
        pack: Any = []
        for node in reversed(self._nodes):
            args = []
            for _ in range(node._n_args):
                args.append(pack.pop())
            if isinstance(node, FunctionalNode):
                pack.append(node._value(*args))
            else:
                pack.append(node._value)
        return pack[0]

    def copy(self) -> Tree:
        return Tree(self._nodes.copy(), self._n_args.copy())

    def set_terminals(self, **kwargs: Any) -> Tree:
        tree_copy = self.copy()
        for i, node in enumerate(tree_copy._nodes):
            if isinstance(node, TerminalNode):
                for name, value in kwargs.items():
                    if node._name == name:
                        tree_copy._nodes[i] = TerminalNode(value=value, name=node._name)
        return tree_copy

    def subtree_id(self, index: int) -> Tuple[int, int]:
        n_index = find_end_subtree_from_i(np.int64(index), self._n_args)
        return (index, n_index)

    def subtree(self, index: int) -> Tree:
        n_index = find_end_subtree_from_i(np.int64(index), self._n_args)
        new_tree = Tree(self._nodes[index:n_index].copy(), self._n_args[index:n_index].copy())
        return new_tree

    def concat(self, index: int, other_tree: Tree) -> Tree:
        to_return = self.copy()
        left, right = self.subtree_id(index)
        to_return._nodes[left:right] = other_tree._nodes.copy()
        to_return._n_args = np.r_[
            to_return._n_args[:left], other_tree._n_args.copy(), to_return._n_args[right:]
        ]
        return to_return

    def get_common_region(self, other_trees: Union[List, NDArray]) -> Tuple:
        if len(other_trees) == 1:
            to_return = common_region_two_trees(self._n_args, other_trees[0]._n_args)
        else:
            trees = [self] + list(other_trees)
            to_return = common_region(trees)

        return to_return

    def get_args_id(self, index: int) -> NDArray[np.int64]:
        args_id = find_id_args_from_i(np.int64(index), self._n_args)
        return args_id

    def get_levels(self, index: int) -> NDArray[np.int64]:
        return get_levels_tree_from_i(np.int64(index), self._n_args)

    def get_max_level(self) -> np.int64:
        return max(self.get_levels(0))

    def get_graph(self, keep_id: bool = False) -> Dict:
        pack: List[int] = []
        edges = []
        nodes = []
        labels = {}
        for i, node in enumerate(reversed(self._nodes)):
            index = len(self) - i - 1
            if keep_id:
                labels[index] = str(index) + ". " + node._sign[:60]
            else:
                labels[index] = node._sign[:60]

            nodes.append(index)

            for _ in range(node._n_args):
                edges.append((index, len(self) - pack.pop() - 1))
            pack.append(i)

        edges.reverse()
        nodes.reverse()

        levels = self.get_levels(0)
        colors = np.zeros(shape=(len(nodes), 4))
        pos = np.zeros(shape=(len(self), 2))
        for i, lvl_i in enumerate(levels):
            total = 0
            cond = lvl_i == np.array(levels)
            h = 1 / (1 + np.sum(cond))
            arange = np.arange(len(pos))[cond]

            for a_j in arange:
                total += h
                pos[a_j][0] = total

            pos[i][1] = -lvl_i

            if isinstance(self._nodes[i], FunctionalNode):
                colors[i] = FUNCTIONAL_COLOR_CODE
            else:
                colors[i] = TERMINAL_COLOR_CODE

        to_return = {"edges": edges, "labels": labels, "nodes": nodes, "pos": pos, "colors": colors}
        return to_return

    def plot(self, ax: Any = None) -> None:
        import networkx as nx

        graph = self.get_graph(keep_id=False)

        G = nx.Graph()
        G.add_nodes_from(graph["nodes"])
        G.add_edges_from(graph["edges"])

        nx.draw_networkx_nodes(
            G, graph["pos"], node_color=graph["colors"], edgecolors="black", linewidths=0.5, ax=ax
        )
        nx.draw_networkx_edges(G, graph["pos"], style="-", ax=ax)
        nx.draw_networkx_labels(G, graph["pos"], graph["labels"], font_size=10, ax=ax)

    @classmethod
    def full_growing_method(cls, uniset: UniversalSet, max_level: int) -> Tree:
        nodes: List[Union[FunctionalNode, TerminalNode, EphemeralNode]] = []
        levels = []
        n_args = []
        possible_steps = [1]
        previous_levels = [-1]
        level_i = -1
        while len(possible_steps):
            possible_steps[-1] = possible_steps[-1] - 1
            if possible_steps[-1] == 0:
                possible_steps.pop()
                level_i = previous_levels.pop() + 1
            else:
                level_i = previous_levels[-1] + 1
            levels.append(level_i)
            if level_i == max_level:
                nodes.append(uniset._random_terminal_or_ephemeral())
                n_args.append(0)
            else:
                nodes.append(uniset._random_functional())
                n_i = nodes[-1]._n_args
                n_args.append(n_i)
                possible_steps.append(n_i)
                previous_levels.append(level_i)
        tree = cls(nodes, n_args)
        return tree

    @classmethod
    def growing_method(cls, uniset: UniversalSet, max_level: int) -> Tree:
        nodes: List[Union[FunctionalNode, TerminalNode, EphemeralNode]] = []
        levels = []
        n_args = []
        possible_steps = [1]
        previous_levels = [-1]
        level_i = -1
        while len(possible_steps):
            possible_steps[-1] = possible_steps[-1] - 1
            if possible_steps[-1] == 0:
                possible_steps.pop()
                level_i = previous_levels.pop() + 1
            else:
                level_i = previous_levels[-1] + 1
            levels.append(level_i)

            if level_i == max_level:
                nodes.append(uniset._random_terminal_or_ephemeral())
                n_args.append(0)
            elif level_i == 0:
                nodes.append(uniset._random_functional())
                n_i = nodes[-1]._n_args
                n_args.append(n_i)
                possible_steps.append(n_i)
                previous_levels.append(level_i)
            else:
                if uniform(low=0, high=1, size=1)[0] < 0.5:
                    nodes.append(uniset._random_terminal_or_ephemeral())
                else:
                    nodes.append(uniset._random_functional())
                n_i = nodes[-1]._n_args
                n_args.append(n_i)

                if n_i > 0:
                    possible_steps.append(n_i)
                    previous_levels.append(level_i)
        tree = cls(nodes, n_args)
        return tree

    @classmethod
    def random_tree(cls, uniset: UniversalSet, max_level: int) -> Tree:
        if uniform(low=0, high=1, size=1)[0] < 0.5:
            tree = cls.full_growing_method(uniset, max_level)
        else:
            tree = cls.growing_method(uniset, max_level)
        return tree


def init_symbolic_regression_uniset(
    X: NDArray[np.float64],
    functional_set_names: Tuple[str, ...] = ("cos", "sin", "add", "sub", "mul", "div"),
    ephemeral_node_generators: Optional[Tuple[Callable, ...]] = None,
):
    SYMBOLIC_FUNCTION_NAME = {
        "cos": create_operator("cos({})", "cos", "cos", cos),
        "sin": create_operator("sin({})", "sin", "sin", sin),
        "add": create_operator("({} + {})", "add", "+", add),
        "sub": create_operator("({} - {})", "sub", "-", sub),
        "mul": create_operator("({} * {})", "mul", "*", mul),
        "div": create_operator("({} / {})", "div", "/", save_div),
        "abs": create_operator("abs({})", "abs", "abs", abs),
        "logabs": create_operator("log(abs({}))", "log(abs)", "log(abs)", logabs),
        "exp": create_operator("exp({})", "exp", "exp", save_exp),
        "logabs": create_operator("sqrt(abs({}))", "sqrt(abs)", "sqrt(abs)", sqrtabs),
    }

    for func_name in functional_set_names:
        if func_name not in SYMBOLIC_FUNCTION_NAME:
            raise ValueError(
                f"Invalid function name '{func_name}'. Available values are: {', '.join(SYMBOLIC_FUNCTION_NAME.keys())}"
            )

    uniset: UniversalSet
    terminal_set: Union[
        List[Union[TerminalNode, EphemeralNode]], Tuple[Union[TerminalNode, EphemeralNode]]
    ]
    functional_set: Union[
        List[Union[TerminalNode, EphemeralNode]], Tuple[Union[TerminalNode, EphemeralNode]]
    ] = []
    n_dimension: int = X.shape[1]

    for functional_name in functional_set_names:
        function = SYMBOLIC_FUNCTION_NAME[functional_name]
        functional_set.append(FunctionalNode(function))

    terminal_set = [TerminalNode(X[:, i], f"x{i}") for i in range(n_dimension)]
    if ephemeral_node_generators is not None:
        for generator in ephemeral_node_generators:
            terminal_set.append(EphemeralNode(generator))

    uniset = UniversalSet(tuple(functional_set), tuple(terminal_set))
    return uniset


def init_net_uniset(
    n_variables: int, input_block_size: int, max_hidden_block_size: int, offset: bool = True
):

    if offset:
        n_dimension = n_variables - 1
    else:
        n_dimension = n_variables

    cut_id: NDArray[np.int64] = np.arange(
        input_block_size, n_dimension, input_block_size, dtype=np.int64
    )

    variables_pool: List = np.split(np.arange(n_dimension), cut_id)
    functional_set = (
        FunctionalNode(create_operator("({} > {})", "gt", ">", gt)),
        FunctionalNode(create_operator("({} + {})", "add", "+", add)),
    )

    def random_hidden_block() -> HiddenBlock:
        return HiddenBlock(max_hidden_block_size)

    terminal_set: List[Union[TerminalNode, EphemeralNode]] = [
        TerminalNode(set(variables), "in{}".format(i)) for i, variables in enumerate(variables_pool)
    ]

    if offset:
        terminal_set.append(
            TerminalNode(value={(n_dimension)}, name="in{}".format(len(variables_pool)))
        )
    terminal_set.append(EphemeralNode(random_hidden_block))

    uniset = UniversalSet(functional_set, tuple(terminal_set))
    return uniset



from __future__ import annotations

from typing import Optional
from typing import Union

import numpy as np
from numpy.typing import NDArray

from ..utils import if_single_or_array_to_float_array
from ..utils import if_single_or_array_to_int_array


def minmax_scale(data: NDArray[Union[np.int64, np.float64]]) -> NDArray[np.float64]:
    """
    Scale the values of a NumPy array between 0 and 1.

    Parameters
    ----------
    data : NDArray[Union[np.int64, np.float64]]
        Input array containing numerical values to be scaled.

    Returns
    -------
    NDArray[np.float64]
        Scaled array with values between 0 and 1.

    Notes
    -----
    This function scales the values of the input array between 0 and 1 using min-max scaling.
    If the minimum and maximum values in the array are equal, the function returns an array of ones.

    Examples
    --------
    >>> import numpy as np
    >>> from thefittest.utils.transformations import minmax_scale
    >>>
    >>> # Example data
    >>> example_data = np.array([2, 5, 10, 8, 3], dtype=np.int64)
    >>>
    >>> # Scale the data using the minmax_scale function
    >>> scaled_data = minmax_scale(example_data)
    >>>
    >>> # Display original and scaled data
    >>> print("Original Data:", example_data)
    Original Data: [ 2  5 10  8  3]
    >>> print("Scaled Data:", scaled_data)
    Scaled Data: [0.    0.375 1.    0.75  0.125]
    """
    data_copy = data.copy()
    max_value = data_copy.max()
    min_value = data_copy.min()
    if max_value == min_value:
        scaled_data = np.ones_like(data_copy, dtype=np.float64)
    else:
        scaled_data = ((data_copy - min_value) / (max_value - min_value)).astype(np.float64)
    return scaled_data


class SamplingGrid:
    """
    SamplingGrid class for transforming populations between binary and floating-point representations.

    This class provides functionality to fit, transform, and inverse transform populations using a specified
    sampling grid. The grid is defined by the left and right borders for each variable, and either the step size (h)
    or the number of bits for each variable.

    Attributes
    ----------
    _left_border : NDArray[np.float64]
        Left border values for each variable in the sampling grid.
    _right_border : NDArray[np.float64]
        Right border values for each variable in the sampling grid.
    _num_variables : int
        Number of variables in the sampling grid.
    _h_per_variable : NDArray[np.float64]
        Step size for each variable in the sampling grid.
    _bits_per_variable : NDArray[np.int64]
        Number of bits for each variable in the sampling grid.
    _reversed_powers : NDArray[np.int64]
        Reversed powers of 2 used for converting binary representations to integers.

    Examples
    --------
    >>> import numpy as np
    >>> from thefittest.utils.transformations import SamplingGrid
    >>>
    >>> # Fit the sampling grid
    >>> grid = SamplingGrid()
    >>> grid.fit(left_border=-5.0, right_border=5.0, num_variables=3, h_per_variable=0.1)
    <thefittest.utils.transformations.SamplingGrid object at ...>
    >>> # Generate a binary population
    >>> string_length = grid.get_bits_per_variable().sum()
    >>> binary_population = np.random.randint(2, size=(5, string_length), dtype=np.int8)
    >>> # Transform the binary population to a floating-point array
    >>> print("Transformed Population:", grid.transform(binary_population))
    Transformed Population:...

    >>> import numpy as np
    >>> from thefittest.utils.transformations import SamplingGrid
    >>>
    >>> # Fit the sampling grid
    >>> grid = SamplingGrid()
    >>> grid.fit(left_border=0.0, right_border=1.0, num_variables=3, h_per_variable=0.1)
    <thefittest.utils.transformations.SamplingGrid object at ...>
    >>>
    >>> # Generate a floating-point population
    >>> floating_population = np.random.rand(5, 3)
    >>> print("Floating-point Population:", floating_population)
    Floating-point Population:...
    >>>
    >>> # Inverse transform the floating-point population to a binary array
    >>> print("Inverse Transformed Population:", grid.inverse_transform(floating_population))
    Inverse Transformed Population: ...

    Methods
    -------
    fit(
        left_border: Union[float, NDArray[np.float64]],
        right_border: Union[float, NDArray[np.float64]],
        num_variables: int,
        h_per_variable: Optional[Union[float, NDArray[np.float64]]] = None,
        bits_per_variable: Optional[Union[int, NDArray[np.int64]]] = None,
    ) -> "SamplingGrid":
        Fit the sampling grid using specified parameters.

    get_left_border() -> NDArray[np.float64]:
        Get the left border values for each variable.

    get_right_border() -> NDArray[np.float64]:
        Get the right border values for each variable.

    get_num_variables() -> int:
        Get the number of variables.

    get_h_per_variable() -> NDArray[np.float64]:
        Get the step size values for each variable.

    get_bits_per_variable() -> NDArray[np.int64]:
        Get the number of bits per variable.

    transform(population: NDArray[np.int8]) -> NDArray[np.float64]:
        Transform a binary population into a floating-point array based on the SamplingGrid parameters.

    inverse_transform(population: NDArray[np.float64]) -> NDArray[np.int8]:
        Inverse transform a floating-point population into a binary array based on the SamplingGrid parameters.
    """

    def __init__(
        self,
    ) -> None:
        self._left_border: NDArray[np.float64]
        self._right_border: NDArray[np.float64]
        self._num_variables: int
        self._h_per_variable: NDArray[np.float64]
        self._bits_per_variable: NDArray[np.int64]
        self._reversed_powers: NDArray[np.int64]

    def get_left_border(self) -> NDArray[np.float64]:
        """
        Get the left border values for each variable.

        Returns
        -------
        NDArray[np.float64]
            Left border values for each variable.
        """
        return self._left_border

    def get_right_border(self) -> NDArray[np.float64]:
        """
        Get the right border values for each variable.

        Returns
        -------
        NDArray[np.float64]
            Right border values for each variable.
        """
        return self._right_border

    def get_num_variables(self) -> int:
        """
        Get the number of variables.

        Returns
        -------
        int
            Number of variables.
        """
        return self._num_variables

    def get_h_per_variable(self) -> NDArray[np.float64]:
        """
        Get the step size values for each variable.

        Returns
        -------
        NDArray[np.float64]
            Step size values for each variable.
        """
        return self._h_per_variable

    def get_bits_per_variable(self) -> NDArray[np.int64]:
        """
        Get the number of bits per variable.

        Returns
        -------
        NDArray[np.int64]
            Number of bits per variable.
        """
        return self._bits_per_variable
    
    def get_str_len(self) -> np.int64:
        """
        Get the total number of bits required to represent the entire population in binary.

        This method sums the number of bits for each variable, providing the total length 
        of the binary string needed to represent a single individual in the population.

        Returns
        -------
        np.int64
            Total number of bits required for binary representation of an individual.
        """
        return self._bits_per_variable.sum()
    
    def fit(
        self,
        left_border: Union[float, NDArray[np.float64]],
        right_border: Union[float, NDArray[np.float64]],
        num_variables: int,
        h_per_variable: Optional[Union[float, NDArray[np.float64]]] = None,
        bits_per_variable: Optional[Union[int, NDArray[np.int64]]] = None,
    ):
        """
        Fit the sampling grid using specified parameters.

        Parameters
        ----------
        left_border : Union[float, NDArray[np.float64]]
            Left border values for each variable.
        right_border : Union[float, NDArray[np.float64]]
            Right border values for each variable.
        num_variables : int
            Number of variables.
        h_per_variable : Optional[Union[float, NDArray[np.float64]]], optional
            Step size values for each variable. Either `h_per_variable` or `bits_per_variable` should be provided.
        bits_per_variable : Optional[Union[int, NDArray[np.int64]]], optional
            Number of bits per variable. Either `h_per_variable` or `bits_per_variable` should be provided.

        Returns
        -------
        SamplingGrid
            The fitted SamplingGrid instance.

        Raises
        ------
        AssertionError
            If both `h_per_variable` and `bits_per_variable` are provided or if neither is provided.

        Notes
        -----
        This method fits the sampling grid using the specified parameters. If `h_per_variable` is provided,
        it calculates the corresponding `bits_per_variable`. If `bits_per_variable` is provided, it calculates
        the corresponding `h_per_variable`. The powers of 2 used for conversion are also calculated and stored.

        Examples
        --------
        >>> import numpy as np
        >>> from thefittest.utils.transformations import SamplingGrid
        >>>
        >>> grid = SamplingGrid()
        >>> grid.fit(left_border=0.0, right_border=1.0, num_variables=3, h_per_variable=0.1)
        <thefittest.utils.transformations.SamplingGrid object at ...>
        >>> print("Grid Left Border:", grid.get_left_border())
        Grid Left Border: [0. 0. 0.]
        >>> print("Grid Right Border:", grid.get_right_border())
        Grid Right Border: [1. 1. 1.]
        >>> print("Number of Variables:", grid.get_num_variables())
        Number of Variables: 3
        >>> print("Step Size per Variable:", grid.get_h_per_variable())
        Step Size per Variable: [0.06666667 0.06666667 0.06666667]
        >>> print("Bits per Variable:", grid.get_bits_per_variable())
        Bits per Variable: [4 4 4]
        >>>
        >>> grid = SamplingGrid()
        >>> grid.fit(left_border=-1.0, right_border=1.0, num_variables=2, bits_per_variable=4)
        <thefittest.utils.transformations.SamplingGrid object at ...>
        >>> print("Grid Left Border:", grid.get_left_border())
        Grid Left Border: [-1. -1.]
        >>> print("Grid Right Border:", grid.get_right_border())
        Grid Right Border: [1. 1.]
        >>> print("Number of Variables:", grid.get_num_variables())
        Number of Variables: 2
        >>> print("Step Size per Variable:", grid.get_h_per_variable())
        Step Size per Variable: [0.13333333 0.13333333]
        >>> print("Bits per Variable:", grid.get_bits_per_variable())
        Bits per Variable: [4 4]
        >>>
        >>> grid = SamplingGrid()
        >>> grid.fit(
        ...     left_border=np.array([-1.0, 0.5, -2.0], dtype=np.float64),
        ...     right_border=np.array([1.0, 5.0, 2.0], dtype=np.float64),
        ...     num_variables=3,
        ...     h_per_variable=np.array([0.05, 1.0, 0.1], dtype=np.float64),
        ... )
        <thefittest.utils.transformations.SamplingGrid object at ...>
        >>> print("Grid Left Border:", grid.get_left_border())
        Grid Left Border: [-1.   0.5 -2. ]
        >>> print("Grid Right Border:", grid.get_right_border())
        Grid Right Border: [1. 5. 2.]
        >>> print("Number of Variables:", grid.get_num_variables())
        Number of Variables: 3
        >>> print("Step Size per Variable:", grid.get_h_per_variable())
        Step Size per Variable: [0.03174603 0.64285714 0.06349206]
        >>> print("Bits per Variable:", grid.get_bits_per_variable())
        Bits per Variable: [6 3 6]
        >>>
        >>> grid = SamplingGrid()
        >>> grid.fit(
        ...     left_border=np.array([-3.5, -2.0, 10.0, 0.9], dtype=np.float64),
        ...     right_border=np.array([3.5, 7.0, 25.0, 1.5], dtype=np.float64),
        ...     num_variables=4,
        ...     bits_per_variable=np.array([8, 16, 3, 40], dtype=np.int64),
        ... )
        <thefittest.utils.transformations.SamplingGrid object at ...>
        >>> print("Grid Left Border:", grid.get_left_border())
        Grid Left Border: [-3.5 -2.  10.   0.9]
        >>> print("Grid Right Border:", grid.get_right_border())
        Grid Right Border: [ 3.5  7.  25.   1.5]
        >>> print("Number of Variables:", grid.get_num_variables())
        Number of Variables: 4
        >>> print("Step Size per Variable:", grid.get_h_per_variable())
        Step Size per Variable: [2.74509804e-02 1.37331197e-04 2.14285714e+00 5.45696821e-13]
        >>> print("Bits per Variable:", grid.get_bits_per_variable())
        Bits per Variable: [ 8 16  3 40]
        """
        self._num_variables = num_variables
        self._left_border = if_single_or_array_to_float_array(left_border, self._num_variables)
        self._right_border = if_single_or_array_to_float_array(right_border, self._num_variables)

        if bits_per_variable is None and h_per_variable is not None:
            self._h_per_variable = if_single_or_array_to_float_array(
                h_per_variable, self._num_variables
            )
            self._culc_num_bits_from_h()
            self._culc_h_from_num_bits()
        elif bits_per_variable is not None and h_per_variable is None:
            self._bits_per_variable = if_single_or_array_to_int_array(
                bits_per_variable, self._num_variables
            )
            self._culc_h_from_num_bits()
        else:
            raise ValueError(
                "Either bits_per_variable or h_per_variable must be defined, but not both."
            )

        self._powers = 2 ** np.arange(self._bits_per_variable.max(), dtype=np.int64)

        return self

    def _culc_h_from_num_bits(self) -> None:
        """
        Calculate the step size (h) for each variable based on the number of bits.

        This method calculates the step size (h) for each variable using the specified number of bits per variable,
        the left and right borders of the sampling grid, and updates the internal state of the SamplingGrid.

        Notes
        -----
        This method should be called internally during the fitting process after setting the number of bits per variable.
        """
        self._h_per_variable = (self._right_border - self._left_border) / (
            2.0**self._bits_per_variable - 1
        )

    def _culc_num_bits_from_h(self) -> None:
        """
        Calculate the number of bits for each variable based on the step size (h).

        This method calculates the number of bits required for each variable based on the specified step size (h),
        the left and right borders of the sampling grid, and updates the internal state of the SamplingGrid.

        Notes
        -----
        This method should be called internally during the fitting process after setting the step size (h) per variable.
        """
        self._bits_per_variable = np.ceil(
            np.log2((self._right_border - self._left_border) / self._h_per_variable + 1)
        ).astype(int)

    @staticmethod
    def bit_to_int(
        bit_array: NDArray[np.int64], powers: Optional[NDArray[np.int64]] = None
    ) -> NDArray[np.int64]:
        """
        Convert a binary array to an integer array using specified powers.

        Parameters
        ----------
        bit_array : NDArray[np.int64]
            2D array where each row represents a binary number.
        powers : Optional[NDArray[np.int64]], optional
            1D array of powers of 2 corresponding to the binary places. If provided, avoids recalculation.

        Returns
        -------
        NDArray[np.int64]
            1D array representing the integer values converted from binary.

        Examples
        --------
        >>> import numpy as np
        >>> from thefittest.utils.transformations import SamplingGrid
        >>>
        >>> # Example 1: Convert binary array to integer array
        >>> binary_array = np.array([[1, 0, 1], [0, 1, 1]], dtype=np.int64)
        >>> result = SamplingGrid.bit_to_int(binary_array)
        >>> print("Converted Integer Array:", result)
        Converted Integer Array: [5 3]
        >>>
        >>> # Example 2: Convert binary array to integer array with  powers
        >>> custom_powers = np.array([1, 2, 4], dtype=np.int64)
        >>> result_custom_powers = SamplingGrid.bit_to_int(binary_array, powers=custom_powers)
        >>> print("Converted Integer Array (Define Powers):", result_custom_powers)
        Converted Integer Array (Define Powers): [5 3]
        """
        num_bits = bit_array.shape[1]
        if powers is None:
            powers = 2 ** np.arange(num_bits, dtype=np.int64)
        reversed_powers = np.flip(powers[:num_bits])
        int_array = np.dot(bit_array, reversed_powers)
        return int_array

    @staticmethod
    def int_to_bit(
        int_array: NDArray[np.int64], powers: Optional[NDArray[np.int64]] = None
    ) -> NDArray[np.int8]:
        """
        Convert a 1D integer array to a 2D binary array.

        Parameters
        ----------
        int_array : NDArray[np.int64]
            1D array of integers to be converted to binary.
        powers : Optional[NDArray[np.int64]], optional
            1D array of powers of 2 corresponding to the binary places. If provided, avoids recalculation.

        Returns
        -------
        NDArray[np.byte]
            2D binary array converted from the 1D integer array.

        Examples
        --------
        >>> import numpy as np
        >>> from thefittest.utils.transformations import SamplingGrid
        >>>
        >>> # Example 1: Convert one-dimensional integer array to binary array
        >>> integer_array = np.array([5, 3], dtype=np.int64)
        >>> result = SamplingGrid.int_to_bit(integer_array)
        >>> print("Converted Binary Array:", result)
        Converted Binary Array: ...
        >>>
        >>> # Example 2: Convert one-dimensional integer array to binary array with powers
        >>> custom_powers = np.array([1, 2, 4], dtype=np.int64)
        >>> integer_array = np.array([5, 3], dtype=np.int64)
        >>> result_custom_powers = SamplingGrid.int_to_bit(integer_array, powers=custom_powers)
        >>> print("Converted Binary Array (Define Powers):", result_custom_powers)
        Converted Binary Array (Define Powers): ...
        """
        num_bits = int(np.ceil(np.log2(np.max(int_array) + 1)))
        bit_array = np.empty(shape=(int_array.shape[0], num_bits), dtype=np.int8)

        if powers is None:
            powers = 2 ** np.arange(num_bits, dtype=np.int64)

        reversed_powers = np.flip(powers[:num_bits])

        int_array = int_array.astype(np.int64)

        for i, reversed_power_i in enumerate(reversed_powers):
            bit_array[:, i] = np.int8((int_array & reversed_power_i) > 0)
        return bit_array

    def _float_to_bit(
        self, float_array: NDArray[np.float64], left: NDArray[np.float64], h: NDArray[np.float64]
    ) -> NDArray[np.int8]:
        """
        Convert a 1D floating-point array to a 2D binary array based on the SamplingGrid parameters.

        Parameters
        ----------
        float_array : NDArray[np.float64]
            1D floating-point array to be converted to binary.
        left : NDArray[np.float64]
            1D array representing the left boundary of the sampling grid used in the conversion process.
        h : NDArray[np.float64]
            1D array representing the step size of the sampling grid used in the conversion process.

        Returns
        -------
        NDArray[np.byte]
            2D binary array converted from the 1D floating-point array.

        Notes
        -----
        This function internally relies on the `int_to_bit` method to convert floating-point numbers to their binary representation.
        """
        grid_number = (float_array - left) / h
        int_array = np.rint(grid_number)
        bit_array = self.int_to_bit(int_array, self._powers)
        return bit_array

    def _decode(self, bit_array_i: NDArray[np.byte]) -> NDArray[np.int64]:
        """
        Decode a 2D binary array representing a single variable into a 1D integer array.

        Parameters
        ----------
        bit_array_i : NDArray[np.byte]
            2D binary array representing the variable. Each row corresponds to a binary representation.

        Returns
        -------
        NDArray[np.int64]
            1D integer array decoded from the 2D binary representation.

        Notes
        -----
        This function internally relies on the `bit_to_int` method for converting the binary representation
        of a variable to its integer counterpart.

        """
        int_convert = self.bit_to_int(bit_array_i, self._powers)
        return int_convert

    def transform(self, population: NDArray[np.int8]) -> NDArray[np.float64]:
        """
        Transform a binary population into a floating-point array based on the SamplingGrid parameters.

        Parameters
        ----------
        population : NDArray[np.int8]
            Population matrix where each row represents a binary array.

        Returns
        -------
        NDArray[np.float64]
            Floating-point array representing the transformed population.

        Notes
        -----
        This function divides the input population into individual variables, decodes each variable from binary to integer,
        and then calculates the corresponding floating-point values using the SamplingGrid parameters.

        Examples
        --------
        >>> import numpy as np
        >>> from thefittest.utils.transformations import SamplingGrid
        >>>
        >>> # Fit the sampling grid
        >>> grid = SamplingGrid()
        >>> grid.fit(left_border=-5.0, right_border=5.0, num_variables=3, h_per_variable=0.1)
        <thefittest.utils.transformations.SamplingGrid object at ...>
        >>> # Generate a binary population
        >>> string_length = grid.get_bits_per_variable().sum()
        >>> binary_population = np.random.randint(2, size=(5, string_length), dtype=np.int8)
        >>> # Transform the binary population to a floating-point array
        >>> print("Transformed Population:", grid.transform(binary_population))
        Transformed Population: ...
        """
        splits = np.add.accumulate(self._bits_per_variable)
        p_parts = np.split(population, splits[:-1], axis=1)

        int_array = np.array(list(map(self._decode, p_parts))).T
        float_array = (
            self._left_border[np.newaxis, :] + self._h_per_variable[np.newaxis, :] * int_array
        )
        return float_array

    def inverse_transform(self, population: NDArray[np.float64]) -> NDArray[np.int8]:
        """
        Inverse transform a floating-point population into a binary array based on the SamplingGrid parameters.

        Parameters
        ----------
        population : NDArray[np.float64]
            Population matrix where each row represents a floating-point array.

        Returns
        -------
        NDArray[np.int8]
            Binary array representing the inverse transformed population.

        Notes
        -----
        This function encodes each variable from floating-point to binary, and then combines the binary representations
        to form the binary array of the inverse transformed population.

        Examples
        --------
        >>> import numpy as np
        >>> from thefittest.utils.transformations import SamplingGrid
        >>>
        >>> # Fit the sampling grid
        >>> grid = SamplingGrid()
        >>> grid.fit(left_border=0.0, right_border=1.0, num_variables=3, h_per_variable=0.1)
        <thefittest.utils.transformations.SamplingGrid object at ...>
        >>>
        >>> # Generate a floating-point population
        >>> floating_population = np.random.rand(5, 3)
        >>> print("Floating-point Population:", floating_population)
        Floating-point Population: ...
        >>>
        >>> # Inverse transform the floating-point population to a binary array
        >>> print("Inverse Transformed Population:", grid.inverse_transform(floating_population))
        Inverse Transformed Population: ...
        """
        map_ = map(self._float_to_bit, population.T, self._left_border, self._h_per_variable)
        bit_array = np.hstack(list(map_))
        return bit_array


class GrayCode(SamplingGrid):
    """
    GrayCode class for transforming populations between gray code and floating-point representations.

    This class extends the functionality of the SamplingGrid for gray code transformations. Gray code is a binary numeral
    system where two successive values differ in only one bit. GrayCode provides methods to convert between binary and gray
    code representations, as well as transforming populations between gray code and floating-point representations using
    the specified sampling grid.

    Attributes
    ----------
    Inherits attributes from SamplingGrid.

    Methods
    -------
    gray_to_bit(gray_array: NDArray[np.byte]) -> NDArray[np.byte]:
        Convert a gray code array to a binary array.

    bit_to_gray(bit_array: NDArray[np.byte]) -> NDArray[np.byte]:
        Convert a binary array to a gray code array.

    _decode(gray_array_i: NDArray[np.byte]) -> NDArray[np.int64]:
        Decode a 2D gray code array representing a single variable into a 1D integer array.

    _float_to_bit(float_array: NDArray[np.float64], left: NDArray[np.float64], h: NDArray[np.float64]) -> NDArray[np.byte]:
        Convert a 1D floating-point array to a 2D gray code array based on the SamplingGrid parameters.

    Examples
    --------
    >>> import numpy as np
    >>> from thefittest.utils.transformations import GrayCode
    >>>
    >>> # Fit the sampling grid with gray code transformation
    >>> grid = GrayCode()
    >>> grid.fit(left_border=-5.0, right_border=5.0, num_variables=3, h_per_variable=0.1)
    <thefittest.utils.transformations.GrayCode object at ...>
    >>> # Generate a binary population using gray code
    >>> string_length = grid.get_bits_per_variable().sum()
    >>> gray_population = np.random.randint(2, size=(5, string_length), dtype=np.byte)
    >>> print("Gray Code Population:", gray_population)
    Gray Code Population: ...
    >>> # Transform the gray code population to a floating-point array
    >>> print("Transformed Population:", grid.transform(gray_population))
    Transformed Population: ...

    >>> # Generate a floating-point population
    >>> floating_population = np.random.rand(5, 3)
    >>> print("Floating-point Population:", floating_population)
    Floating-point Population: ...
    >>>
    >>> # Inverse transform the floating-point population to a gray code array
    >>> inverse_transformed_population = grid.inverse_transform(floating_population)
    >>> print("Inverse Transformed Population (Gray Code):", inverse_transformed_population)
    Inverse Transformed Population (Gray Code): ...
    """

    def __init__(self) -> None:
        SamplingGrid.__init__(self)

    @staticmethod
    def gray_to_bit(gray_array: NDArray[np.byte]) -> NDArray[np.byte]:
        """
        Convert a gray code array to a binary array.

        Parameters
        ----------
        gray_array : NDArray[np.byte]
            2D array where each row represents a gray code number.

        Returns
        -------
        NDArray[np.byte]
            2D binary array converted from the gray code array.

        Examples
        --------
        >>> import numpy as np
        >>> from thefittest.utils.transformations import GrayCode
        >>>
        >>> # Example: Convert gray code array to binary array using GrayCode.gray_to_bit method
        >>> gray_array = np.array([[1, 0, 1], [0, 1, 0]], dtype=np.byte)
        >>> result = GrayCode.gray_to_bit(gray_array)
        >>> print("Converted Binary Array:", result)
        Converted Binary Array: ...
        """
        bit_array = np.logical_xor.accumulate(gray_array, axis=-1).astype(np.byte)
        return bit_array

    @staticmethod
    def bit_to_gray(bit_array: NDArray[np.byte]) -> NDArray[np.byte]:
        """
        Convert a binary array to a gray code array.

        Parameters
        ----------
        bit_array : NDArray[np.byte]
            2D binary array where each row represents a binary number.

        Returns
        -------
        NDArray[np.byte]
            2D gray code array converted from the binary array.

        Examples
        --------
        >>> import numpy as np
        >>> from thefittest.utils.transformations import GrayCode
        >>>
        >>> # Example: Convert binary array to gray code array using GrayCode.bit_to_gray method
        >>> binary_array = np.array([[1, 0, 1], [0, 1, 0]], dtype=np.byte)
        >>> result = GrayCode.bit_to_gray(binary_array)
        >>> print("Converted Gray Code Array:", result)
        Converted Gray Code Array: ...
        """
        cut_gray = np.logical_xor(bit_array[:, :-1], bit_array[:, 1:])
        gray_array = np.hstack([bit_array[:, 0].reshape(-1, 1), cut_gray])
        return gray_array

    def _decode(self, gray_array_i: NDArray[np.byte]) -> NDArray[np.int64]:
        """
        Decode a 2D gray code array representing a single variable into a 1D integer array.

        Parameters
        ----------
        gray_array_i : NDArray[np.byte]
            2D gray code array representing the variable. Each row corresponds to a gray code representation.

        Returns
        -------
        NDArray[np.int64]
            1D integer array decoded from the 2D gray code representation.

        """
        bit_array_i = self.gray_to_bit(gray_array_i)
        int_convert = self.bit_to_int(bit_array_i, self._powers)
        return int_convert

    def _float_to_bit(
        self, float_array: NDArray[np.float64], left: NDArray[np.float64], h: NDArray[np.float64]
    ) -> NDArray[np.byte]:
        """
        Convert a 1D floating-point array to a 2D gray code array based on the GrayCode parameters.

        Parameters
        ----------
        float_array : NDArray[np.float64]
            1D floating-point array to be converted to gray code.
        left : NDArray[np.float64]
            1D array representing the left boundary of the sampling grid used in the conversion process.
        h : NDArray[np.float64]
            1D array representing the step size of the sampling grid used in the conversion process.

        Returns
        -------
        NDArray[np.byte]
            2D gray code array converted from the 1D floating-point array.
        """
        grid_number = (float_array - left) / h
        int_array = np.rint(grid_number)
        bit_array = self.int_to_bit(int_array)
        gray_array = self.bit_to_gray(bit_array)
        return gray_array



from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Union

import numpy as np
from numpy.typing import NDArray

from ._geneticalgorithm import GeneticAlgorithm
from ..base import Tree
from ..base import UniversalSet
from ..utils.random import randint


class GeneticProgramming(GeneticAlgorithm):
    """Koza, John R.. “Genetic programming - on the programming of computers by means
    of natural selection.” Complex Adaptive Systems (1993)"""

    def __init__(
        self,
        fitness_function: Callable[[NDArray[Any]], NDArray[np.float64]],
        uniset: UniversalSet,
        iters: int,
        pop_size: int,
        tour_size: int = 2,
        mutation_rate: float = 0.05,
        parents_num: int = 7,
        elitism: bool = True,
        selection: str = "rank",
        crossover: str = "gp_standard",
        mutation: str = "gp_weak_grow",
        max_level: int = 16,
        init_level: int = 5,
        init_population: Optional[NDArray] = None,
        genotype_to_phenotype: Optional[Callable[[NDArray[Any]], NDArray[Any]]] = None,
        optimal_value: Optional[float] = None,
        termination_error_value: float = 0.0,
        no_increase_num: Optional[int] = None,
        minimization: bool = False,
        show_progress_each: Optional[int] = None,
        keep_history: bool = False,
        n_jobs: int = 1,
        fitness_function_args: Optional[Dict] = None,
        genotype_to_phenotype_args: Optional[Dict] = None,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        on_generation: Optional[Callable] = None,
    ):
        GeneticAlgorithm.__init__(
            self,
            fitness_function=fitness_function,
            iters=iters,
            pop_size=pop_size,
            str_len=1,
            tour_size=tour_size,
            mutation_rate=mutation_rate,
            parents_num=parents_num,
            elitism=elitism,
            selection=selection,
            crossover=crossover,
            mutation=mutation,
            init_population=init_population,
            genotype_to_phenotype=genotype_to_phenotype,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history,
            n_jobs=n_jobs,
            fitness_function_args=fitness_function_args,
            genotype_to_phenotype_args=genotype_to_phenotype_args,
            random_state=random_state,
            on_generation=on_generation,
        )

        self._uniset: UniversalSet = uniset
        self._max_level: int = max_level
        self._init_level: int = init_level

    @staticmethod
    def half_and_half(pop_size: int, uniset: UniversalSet, max_level: int) -> NDArray:
        level = randint(2, max_level, 1)[0]
        population = [Tree.random_tree(uniset, level) for _ in range(pop_size)]
        population_numpy = np.array(population, dtype=object)
        return population_numpy

    def _first_generation(self: GeneticProgramming) -> None:
        if self._init_population is None:
            self._population_g_i = self._population_g_i = self.half_and_half(
                pop_size=self._pop_size, uniset=self._uniset, max_level=self._init_level
            )
        else:
            self._population_g_i = self._init_population.copy()

    def _get_new_individ_g(
        self: GeneticProgramming,
        specified_selection: str,
        specified_crossover: str,
        specified_mutation: str,
    ) -> Tree:
        selection_func, tour_size = self._selection_pool[specified_selection]
        crossover_func, quantity = self._crossover_pool[specified_crossover]
        mutation_func, proba, is_constant_rate = self._mutation_pool[specified_mutation]

        selected_id = selection_func(
            self._fitness_scale_i, self._fitness_rank_i, np.int64(tour_size), np.int64(quantity)
        )

        offspring_no_mutated = crossover_func(
            self._population_g_i[selected_id],
            self._fitness_scale_i[selected_id],
            self._fitness_rank_i[selected_id],
            self._max_level,
        )

        if is_constant_rate:
            proba = proba
        else:
            proba = proba / len(offspring_no_mutated)

        offspring = mutation_func(offspring_no_mutated, self._uniset, proba, self._max_level)
        return offspring



from typing import Any
from typing import Callable
from typing import Tuple
from typing import List
from typing import Union

from numba import boolean
from numba import float64
from numba import int64
from numba import njit
from numba.typed import Dict as numbaDict
from numba.types import List as numbaListType

import numpy as np
from numpy.typing import ArrayLike
from numpy.typing import NDArray

import torch
from torch.optim import Optimizer as TorchOptimizer

MIN_VALUE = np.finfo(np.float64).min
MAX_VALUE = np.finfo(np.float64).max


def common_region(trees: Union[List, NDArray]) -> Tuple:
    terminate = False
    indexes = []
    common_indexes: List[List[int]] = []
    border_indexes: List[List[int]] = []
    for tree in trees:
        indexes.append(list(range(len(tree._nodes))))
        common_indexes.append([])
        border_indexes.append([])

    while not terminate:
        inner_break = False
        iters = min(list(map(len, indexes)))

        for i in range(iters):
            first_n_args = trees[0]._nodes[indexes[0][i]]._n_args
            common_indexes[0].append(indexes[0][i])
            for j in range(1, len(indexes)):
                common_indexes[j].append(indexes[j][i])
                if first_n_args != trees[j]._nodes[indexes[j][i]]._n_args:
                    inner_break = True

            if inner_break:
                for j in range(0, len(indexes)):
                    border_indexes[j].append(indexes[j][i])
                break

        for j in range(len(indexes)):
            _, right = trees[j].subtree_id(common_indexes[j][-1])
            delete_to = indexes[j].index(right - 1) + 1
            indexes[j] = indexes[j][delete_to:]

            if len(indexes[j]) < 1:
                terminate = True
                break

    return common_indexes, border_indexes


def common_region_two_trees(
    n_args_array_1: NDArray[np.int64], n_args_array_2: NDArray[np.int64]
) -> Tuple:
    index_list_1 = range(len(n_args_array_1))
    index_list_2 = range(len(n_args_array_2))
    index_1 = np.int64(0)
    index_2 = np.int64(0)

    common_1: List[List[int]] = []
    common_2: List[List[int]] = []
    border_1 = []
    border_2 = []

    while True:
        if index_1 < len(n_args_array_1) and index_2 < len(n_args_array_2):
            id_1 = index_1
            id_2 = index_2
            end = find_first_difference_between_two(n_args_array_1[id_1:], n_args_array_2[id_2:])
            index_1 = index_1 + end
            index_2 = index_2 + end
            common_1.extend(index_list_1[id_1 : index_1 + 1])
            common_2.extend(index_list_2[id_2 : index_2 + 1])

        if len(n_args_array_1) - 1 > index_1 or len(n_args_array_2) - 1 > index_2:
            border_1.append(index_1)
            border_2.append(index_2)
            index_1 = find_end_subtree_from_i(index_1, n_args_array_1)
            index_2 = find_end_subtree_from_i(index_2, n_args_array_2)
        else:
            break

    return [common_1, common_2], [border_1, border_2]


@njit(int64(int64, int64[:]))
def find_end_subtree_from_i(index: np.int64, n_args_array: NDArray[np.int64]) -> np.int64:
    n_index = index + 1
    possible_steps = n_args_array[index]
    while possible_steps:
        possible_steps += n_args_array[n_index] - 1
        n_index += 1
    return n_index


@njit(int64[:](int64, int64[:]))
def find_id_args_from_i(index: np.int64, n_args_array: NDArray[np.int64]) -> NDArray[np.int64]:
    out = np.empty(n_args_array[index], dtype=np.int64)
    out[0] = index + 1
    for i in range(1, len(out)):
        next_stop = find_end_subtree_from_i(out[i - 1], n_args_array)
        out[i] = next_stop
    return out


@njit(int64[:](int64, int64[:]))
def get_levels_tree_from_i(origin: np.int64, n_args_array: NDArray[np.int64]) -> NDArray[np.int64]:
    d_i = -1
    s = [1]
    d = [-1]
    result_list = []
    for n_arg in n_args_array[origin:]:
        s[-1] = s[-1] - 1
        if s[-1] == 0:
            s.pop()
            d_i = d.pop() + 1
        else:
            d_i = d[-1] + 1
        result_list.append(d_i)
        if n_arg > 0:
            s.append(n_arg)
            d.append(d_i)
        if len(s) == 0:
            break
    return np.array(result_list, dtype=np.int64)


@njit(int64(int64[:], int64[:]))
def find_first_difference_between_two(
    array_1: NDArray[np.int64], array_2: NDArray[np.int64]
) -> np.int64:
    for i in np.arange(min(len(array_1), len(array_2)), dtype=np.int64):
        if array_1[i] != array_2[i]:
            break
    return i


@njit(int64(float64, float64[:]))
def binary_search_interval(value: np.float64, intervals: NDArray[np.float64]) -> int:
    if value <= intervals[0]:
        ind = 0
    else:
        left = 0
        right = len(intervals) - 1
        while right - left > 1:
            mid = (left + right) // 2
            if value <= intervals[mid]:
                right = mid
            else:
                left = mid
        ind = right
    return ind


@njit(boolean(int64, int64[:], int64))
def check_for_value(value: np.int64, index_array: NDArray[np.int64], end: np.int64) -> bool:
    found = False
    for i in range(end):
        if value == index_array[i]:
            found = True
            break
    return found


@njit(int64[:](float64[:], int64))
def argsort_k(array: NDArray[np.float64], k: np.int64) -> NDArray[np.int64]:
    size = len(array)
    array_copy = array.copy()
    to_return = np.arange(size, dtype=np.int64)
    for i in range(k):
        max_ = array_copy[i]
        for j in range(i, size):
            if array_copy[j] > max_:
                max_ = array_copy[j]
                max_id = j
        array_copy[i], array_copy[max_id] = array_copy[max_id], array_copy[i]
        to_return[i], to_return[max_id] = to_return[max_id], to_return[i]
    return to_return


@njit(int64[:](float64[:], float64))
def find_pbest_id(array: NDArray[np.float64], p: np.float64) -> NDArray[np.int64]:
    size = len(array)
    count = max(np.int64(1), np.int64(p * size))
    argsort = argsort_k(array, count)
    to_return = argsort[:count]
    return to_return


@njit(float64[:, :](float64[:, :]))
def max_axis(array: NDArray[np.float64]) -> NDArray[np.float64]:
    res = np.zeros((array.shape[0], 1), dtype=np.float64)
    for i in range(array.shape[0]):
        res[i] = np.max(array[i])
    return res


@njit(int64[:](int64[:, :]))
def voting(argmax_y: NDArray[np.int64]) -> NDArray[np.int64]:
    winner = np.empty(argmax_y.shape[1], dtype=np.int64)
    for i in range(argmax_y.shape[1]):
        counter = numbaDict.empty(key_type=int64, value_type=int64)

        for argmax_y_i_j in argmax_y[:, i]:
            if argmax_y_i_j not in counter:
                counter[argmax_y_i_j] = 1
            else:
                counter[argmax_y_i_j] += 1

        max_key = -1
        max_value = 0
        for key, value in counter.items():
            if value > max_value:
                max_value = value
                max_key = key

        winner[i] = max_key
    return winner


@njit(float64[:, :](float64[:, :]))
def softmax_numba(X: NDArray[np.float64]) -> NDArray[np.float64]:
    exps = np.exp(X - max_axis(X))
    sum_ = np.sum(exps, axis=1)
    for j in range(sum_.shape[0]):
        if sum_[j] == 0:
            sum_[j] = 1
    result = ((exps).T / sum_).T
    return result


@njit(float64[:, :](float64[:, :], int64))
def multiactivation2d(X: NDArray[np.float64], activ_id: np.int64) -> NDArray[np.float64]:
    if activ_id == 0:
        result = 1 / (1 + np.exp(-X))
    elif activ_id == 1:
        result = X * (X > 0)
    elif activ_id == 2:
        result = np.exp(-(X**2))
    elif activ_id == 3:
        result = np.tanh(X)
    elif activ_id == 4:
        result = X
    elif activ_id == 5:
        result = softmax_numba(X)
    return result


@njit
def forward(
    weights: NDArray[np.float64],
    nodes: NDArray[np.float64],
    from_: numbaListType(NDArray[np.int64]),
    to_: numbaListType(NDArray[np.int64]),
    weights_id: numbaListType(NDArray[np.int64]),
    activs_code: numbaListType(NDArray[np.int64]),
    activs_nodes: numbaListType(numbaListType(NDArray[np.int64])),
) -> NDArray[np.float64]:
    for from_i, to_i, weights_id_i, a_code_i, a_nodes_i in zip(
        from_, to_, weights_id, activs_code, activs_nodes
    ):
        arr_mask = weights[weights_id_i.flatten()]
        weights_i = arr_mask.reshape(weights_id_i.shape)
        out = np.dot(nodes[from_i].T, weights_i.T)
        nodes[to_i] = out.T

        for a_code_i_i, a_nodes_i_i in zip(a_code_i, a_nodes_i):
            nodes[a_nodes_i_i] = multiactivation2d(nodes[a_nodes_i_i].T, a_code_i_i).T

    return nodes


@njit
def forward2d(
    X: NDArray[np.float64],
    inputs: NDArray[np.int64],
    n_hiddens: np.int64,
    outputs: NDArray[np.int64],
    from_: numbaListType(NDArray[np.int64]),
    to_: numbaListType(NDArray[np.int64]),
    weights_id: numbaListType(NDArray[np.int64]),
    activs_code: numbaListType(NDArray[np.int64]),
    activs_nodes: numbaListType(numbaListType(NDArray[np.int64])),
    weights: NDArray[np.float64],
) -> NDArray[np.float64]:
    outs = np.empty(shape=(len(weights), X.shape[0], len(outputs)))
    num_nodes = X.shape[1] + n_hiddens + len(outputs)
    shape = (num_nodes, len(X))
    nodes = np.empty(shape, dtype=np.float64)
    nodes[inputs] = X.T[inputs]

    for n in range(outs.shape[0]):
        forward(weights[n], nodes, from_, to_, weights_id, activs_code, activs_nodes)

        outs[n] = nodes[outputs].T
    return outs


def numpy_group_by(group: NDArray, by: NDArray) -> Tuple:
    argsort = np.argsort(by)
    group = group[argsort]
    by = by[argsort]

    keys, cut_index = np.unique(by, return_index=True)
    groups = np.split(group, cut_index)[1:]
    return keys, groups


def array_like_to_numpy_X_y(
    X: ArrayLike, y: ArrayLike, y_numeric=True
) -> Tuple[NDArray[np.float64]]:
    X = np.array(X, dtype=np.float64)
    if y_numeric:
        y = np.array(y, dtype=np.float64)
    else:
        y = np.array(y, dtype=np.int64)
    return X, y


def save_div(x: Union[float, np.ndarray], y: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
    result: Union[float, np.ndarray]
    if isinstance(y, np.ndarray):
        result = np.divide(x, y, out=np.ones_like(y, dtype=np.float64), where=y != 0)
    else:
        if y == 0:
            result = 0.0
        else:
            result = x / y
    result = np.clip(result, MIN_VALUE, MAX_VALUE)
    return result


def logabs(y: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
    y_ = np.abs(y)
    if isinstance(y_, np.ndarray):
        result = np.log(y_, out=np.ones_like(y_, dtype=np.float64), where=y_ != 0)
    else:
        if y_ == 0:
            result = 1
        else:
            result = np.log(y_)
    return result


def save_exp(x: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
    result = np.clip(np.exp(x), MIN_VALUE, MAX_VALUE)
    return result


def sqrtabs(x: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
    result = np.sqrt(np.abs(x))
    return result


def cos(x: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
    return np.cos(x)


def sin(x: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
    return np.sin(x)


class Operator:
    def __init__(self, formula: str, name: str, sign: str, operation: Callable) -> None:
        self._formula = formula
        self.__name__ = name
        self._sign = sign
        self._operation = operation

    def _write(self, *args: Any) -> str:
        formula = self._formula.format(*args)
        return formula

    def __call__(self, *args: Any) -> Any:
        return self._operation(*args)


def create_operator(formula: str, name: str, sign: str, operation: Callable) -> Operator:
    """
    Creates a new mathematical operator for performing a given operation.

    This factory function allows defining a new operator with a custom function
    that performs a mathematical operation. The defined operator can be used to
    perform the operation on scalar values or arrays.

    Parameters
    ----------
    formula : str
        The string representation of the operator's formula, where the operation's arguments
        are denoted by curly braces. For example, "({} + {})" for an addition operation.
    name : str
        The name of the operator, used for identification.
    sign : str
        The symbol of the operator, used for representation in formulas.
    operation : Callable
        The function that performs the operation. It must accept as many arguments as
        indicated by `formula`, and return the result of the operation.

    Returns
    -------
    Operator
        An instance of the Operator class, representing the new mathematical operator.

    Examples
    --------
    Creating an operator to calculate the cosine:

    >>> cos = create_operator("cos({})", "cos", "cos", np.cos)
    >>> print(cos(np.pi))
    -1.0

    Creating an operator for adding two numbers:

    >>> add = create_operator("({} + {})", "add", "+", lambda x, y: x + y)
    >>> print(add(5, 3))
    8

    Creating an operator for subtraction:

    >>> sub = create_operator("({} - {})", "sub", "-", lambda x, y: x - y)
    >>> print(sub(10, 4))
    6
    """
    return Operator(formula, name, sign, operation)


def if_single_or_array_to_float_array(
    single_or_array: Union[float, int, np.number, NDArray[np.number]], num_variables: int
) -> NDArray[np.float64]:
    """
    Convert a single value or an array to a 1D float array.

    Parameters
    ----------
    single_or_array : Union[float, int, np.number, NDArray[np.number]]
        A single numeric value or an array of numeric values.

    Returns
    -------
    NDArray[np.float64]
        1D float array.

    """
    if isinstance(single_or_array, (int, float, np.number)):
        array = np.full(shape=num_variables, fill_value=single_or_array, dtype=np.float64)
    else:
        assert len(single_or_array) == num_variables
        array = single_or_array.astype(np.float64)

    return array


def if_single_or_array_to_int_array(
    single_or_array: Union[float, int, np.number, NDArray[np.number]], num_variables: int
) -> NDArray[np.int64]:
    """
    Convert a single value or an array to a 1D int array.

    Parameters
    ----------
    single_or_array : Union[float, int, np.number, NDArray[np.number]]
        A single numeric value or an array of numeric values.

    Returns
    -------
    NDArray[np.int64]
        1D int array.

    """
    if isinstance(single_or_array, (int, float, np.number)):
        array = np.full(shape=num_variables, fill_value=single_or_array, dtype=np.int64)
    else:
        assert len(single_or_array) == num_variables
        array = single_or_array.astype(np.int64)

    return array


def _is_torch_optimizer(cls: Any) -> bool:
    try:
        return issubclass(cls, TorchOptimizer)
    except TypeError:
        return False



import numpy as np

from sklearn.metrics import accuracy_score as accuracy_score_sklearn
from sklearn.metrics import confusion_matrix as sklearn_confusion_matrix
from sklearn.metrics import f1_score as sklearn_f1_score
from sklearn.metrics import log_loss
from sklearn.metrics import mean_squared_error as mean_squared_error_sklearn
from sklearn.metrics import precision_score as sklearn_precision_score
from sklearn.metrics import r2_score
from sklearn.metrics import recall_score as sklearn_recall_score

from thefittest.utils._metrics import accuracy_score
from thefittest.utils._metrics import categorical_crossentropy
from thefittest.utils._metrics import coefficient_determination
from thefittest.utils._metrics import confusion_matrix
from thefittest.utils._metrics import f1_score
from thefittest.utils._metrics import precision_score
from thefittest.utils._metrics import recall_score


def test_coefficient_determination():
    # Test data
    y_true = np.array([1.0, 2.0, 3.0], dtype=np.float64)
    y_predict = np.array([1.1, 1.9, 3.1], dtype=np.float64)

    # Calculate R-squared using both functions
    r2_njit = coefficient_determination(y_true, y_predict)
    r2_sklearn = r2_score(y_true, y_predict)

    # Check if the results are equal within a small tolerance
    assert np.isclose(r2_njit, r2_sklearn, rtol=1e-10, atol=1e-10)


def test_categorical_crossentropy():
    # Test data
    target = np.array([[0, 1, 0], [1, 0, 0], [0, 0, 1]], dtype=np.float64)
    output = np.array([[0.1, 0.9, 0.0], [0.9, 0.1, 0.0], [0.0, 0.1, 0.9]], dtype=np.float64)

    # Calculate categorical crossentropy using both functions
    crossentropy_njit = categorical_crossentropy(target, output)
    crossentropy_sklearn = log_loss(target, output)

    # Check if the results are equal within a small tolerance
    assert np.isclose(crossentropy_njit, crossentropy_sklearn, rtol=1e-4, atol=1e-4)


def test_accuracy_score():
    # Test data
    y_true = np.array([1, 0, 1, 1, 0, 1, 0, 0, 1, 1], dtype=np.int64)
    y_predict = np.array([1, 0, 1, 0, 0, 1, 0, 1, 1, 1], dtype=np.int64)

    # Calculate accuracy score using both functions
    accuracy_score_njit_result = accuracy_score(y_true, y_predict)
    accuracy_score_sklearn_result = accuracy_score_sklearn(y_true, y_predict)

    # Check if the results are equal within a small tolerance
    assert np.isclose(
        accuracy_score_njit_result, accuracy_score_sklearn_result, rtol=1e-10, atol=1e-10
    )


def test_confusion_matrix():
    # Test data
    y_true = np.array([1, 0, 1, 1, 0, 1, 0, 0, 1, 1], dtype=np.int64)
    y_predict = np.array([1, 0, 1, 0, 0, 1, 0, 1, 1, 1], dtype=np.int64)

    # Calculate confusion matrix using both functions
    confusion_matrix_njit_result = confusion_matrix(y_true, y_predict)
    confusion_matrix_sklearn_result = sklearn_confusion_matrix(y_true, y_predict)

    # Check if the results are equal
    assert np.all(confusion_matrix_njit_result == confusion_matrix_sklearn_result)


def test_recall_score():
    # Test data
    y_true = np.array([1, 0, 1, 1, 0, 1, 0, 0, 1, 1], dtype=np.int64)
    y_predict = np.array([1, 0, 1, 0, 0, 1, 0, 1, 1, 1], dtype=np.int64)

    # Calculate recall scores using both functions
    recall_score_njit_result = recall_score(y_true, y_predict)
    recall_score_sklearn_result = sklearn_recall_score(y_true, y_predict, average="macro")

    # Check if the results are almost equal (considering potential floating-point differences)
    assert np.isclose(recall_score_njit_result, recall_score_sklearn_result, rtol=1e-10, atol=1e-10)


def test_precision_score():
    # Test data
    y_true = np.array([1, 0, 1, 1, 0, 1, 0, 0, 1, 1], dtype=np.int64)
    y_predict = np.array([1, 0, 1, 0, 0, 1, 0, 1, 1, 1], dtype=np.int64)

    # Calculate precision scores using both functions
    precision_score_njit_result = precision_score(y_true, y_predict)
    precision_score_sklearn_result = sklearn_precision_score(y_true, y_predict, average="macro")

    # Check if the results are almost equal (considering potential floating-point differences)
    assert np.isclose(
        precision_score_njit_result, precision_score_sklearn_result, rtol=1e-10, atol=1e-10
    )


def test_f1_score():
    # Test data
    y_true = np.array([1, 0, 1, 1, 0, 1, 0, 0, 1, 1], dtype=np.int64)
    y_predict = np.array([1, 0, 1, 0, 0, 1, 0, 1, 1, 1], dtype=np.int64)

    # Calculate F1 scores using both functions
    f1_score_njit_result = f1_score(y_true, y_predict)
    f1_score_sklearn_result = sklearn_f1_score(y_true, y_predict, average="macro")

    # Check if the results are almost equal (considering potential floating-point differences)
    assert np.isclose(f1_score_njit_result, f1_score_sklearn_result, rtol=1e-10, atol=1e-10)



from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
from numpy.typing import NDArray

from ._selfcga import SelfCGA
from ..base import Tree
from ..utils import numpy_group_by
from ..utils.random import random_sample


class PDPGA(SelfCGA):
    """Niehaus, J., Banzhaf, W. (2001). Adaption of Operator Probabilities in
    Genetic Programming. In: Miller, J., Tomassini, M., Lanzi, P.L., Ryan, C.,
    Tettamanzi, A.G.B., Langdon, W.B. (eds) Genetic Programming. EuroGP 2001.
    Lecture Notes in Computer Science, vol 2038. Springer, Berlin, Heidelberg.
    https://doi.org/10.1007/3-540-45355-5_26
    """

    def __init__(
        self,
        fitness_function: Callable[[NDArray[Any]], NDArray[np.float64]],
        iters: int,
        pop_size: int,
        str_len: int,
        tour_size: int = 2,
        mutation_rate: float = 0.05,
        parents_num: int = 2,
        elitism: bool = True,
        selections: Tuple[str, ...] = (
            "proportional",
            "rank",
            "tournament_3",
            "tournament_5",
            "tournament_7",
        ),
        crossovers: Tuple[str, ...] = (
            "empty",
            "one_point",
            "two_point",
            "uniform_2",
            "uniform_7",
            "uniform_prop_2",
            "uniform_prop_7",
            "uniform_rank_2",
            "uniform_rank_7",
            "uniform_tour_3",
            "uniform_tour_7",
        ),
        mutations: Tuple[str, ...] = ("weak", "average", "strong"),
        init_population: Optional[NDArray[np.byte]] = None,
        genotype_to_phenotype: Optional[Callable[[NDArray[np.byte]], NDArray[Any]]] = None,
        optimal_value: Optional[float] = None,
        termination_error_value: float = 0.0,
        no_increase_num: Optional[int] = None,
        minimization: bool = False,
        show_progress_each: Optional[int] = None,
        keep_history: bool = False,
        n_jobs: int = 1,
        fitness_function_args: Optional[Dict] = None,
        genotype_to_phenotype_args: Optional[Dict] = None,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        on_generation: Optional[Callable] = None,
    ):
        SelfCGA.__init__(
            self,
            fitness_function=fitness_function,
            iters=iters,
            pop_size=pop_size,
            str_len=str_len,
            tour_size=tour_size,
            mutation_rate=mutation_rate,
            parents_num=parents_num,
            elitism=elitism,
            selections=selections,
            crossovers=crossovers,
            mutations=mutations,
            selection_threshold_proba=0.2 / len(selections),
            crossover_threshold_proba=0.2 / len(crossovers),
            mutation_threshold_proba=0.2 / len(mutations),
            init_population=init_population,
            genotype_to_phenotype=genotype_to_phenotype,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history,
            n_jobs=n_jobs,
            fitness_function_args=fitness_function_args,
            genotype_to_phenotype_args=genotype_to_phenotype_args,
            random_state=random_state,
            on_generation=on_generation,
        )

        self._previous_fitness_i: List = []
        self._success_i: NDArray[np.bool_]

    def _choice_parent(self: PDPGA, fitness_i_selected: NDArray[np.float64]) -> np.float64:
        index = random_sample(len(fitness_i_selected), 1, True)[0]
        choosen = fitness_i_selected[index]
        return choosen

    def _get_new_proba_pdp(
        self: PDPGA,
        proba_dict: Dict["str", float],
        operators: NDArray,
        threshold: float,
    ) -> Dict["str", float]:
        n = len(proba_dict)
        operators, r_values = self._culc_r_i(self._success_i, operators, list(proba_dict.keys()))
        scale = np.sum(r_values)
        proba_value = threshold + (r_values * ((1.0 - n * threshold) / scale))
        new_proba_dict = dict(zip(operators, proba_value))
        new_proba_dict = dict(sorted(new_proba_dict.items()))
        return new_proba_dict

    def _culc_r_i(
        self: PDPGA, success: NDArray[np.bool_], operator: NDArray, operator_keys: List[str]
    ) -> Tuple:
        keys, groups = numpy_group_by(group=success, by=operator)
        r_values = np.array(list(map(lambda x: (np.sum(x) ** 2 + 1) / (len(x) + 1), groups)))
        for key in operator_keys:
            if key not in keys:
                keys = np.append(keys, key)
                r_values = np.append(r_values, 0.0)

        return keys, r_values

    def _adapt(self: PDPGA) -> None:
        if len(self._previous_fitness_i):
            self._success_i = np.array(self._previous_fitness_i, dtype=np.float64) < self._fitness_i

            self._selection_proba = self._get_new_proba_pdp(
                self._selection_proba, self._selection_operators, self._thresholds["selection"]
            )

            self._crossover_proba = self._get_new_proba_pdp(
                self._crossover_proba, self._crossover_operators, self._thresholds["crossover"]
            )

            self._mutation_proba = self._get_new_proba_pdp(
                self._mutation_proba, self._mutation_operators, self._thresholds["mutation"]
            )

            self._previous_fitness_i = []

    def _get_new_individ_g(
        self: PDPGA,
        specified_selection: str,
        specified_crossover: str,
        specified_mutation: str,
    ) -> Union[Tree, np.byte]:
        selection_func, tour_size = self._selection_pool[specified_selection]
        crossover_func, quantity = self._crossover_pool[specified_crossover]
        mutation_func, proba, is_constant_rate = self._mutation_pool[specified_mutation]

        selected_id = selection_func(
            self._fitness_scale_i, self._fitness_rank_i, np.int64(tour_size), np.int64(quantity)
        )

        previous_fitness = self._choice_parent(self._fitness_i[selected_id])
        self._previous_fitness_i.append(previous_fitness)

        offspring_no_mutated = crossover_func(
            self._population_g_i[selected_id],
            self._fitness_scale_i[selected_id],
            self._fitness_rank_i[selected_id],
        )

        if is_constant_rate:
            proba = proba
        else:
            proba = proba / len(offspring_no_mutated)

        offspring = mutation_func(offspring_no_mutated, np.float64(proba))
        return offspring



from __future__ import annotations

from functools import partial
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Union

import numpy as np
from numpy.typing import NDArray

from ._differentialevolution import DifferentialEvolution
from ..utils.random import uniform


class jDE(DifferentialEvolution):
    """Brest, Janez & Greiner, Sao & Bošković, Borko & Mernik, Marjan & Zumer, Viljem. (2007).
    Self-Adapting Control Parameters in Differential Evolution: A Comparative Study on Numerical
    Benchmark Problems. Evolutionary Computation, IEEE Transactions on. 10.
      646 - 657. 10.1109/TEVC.2006.872133."""

    def __init__(
        self,
        fitness_function: Callable[[NDArray[Any]], NDArray[np.float64]],
        iters: int,
        pop_size: int,
        left_border: Union[float, int, np.number, NDArray[np.number]],
        right_border: Union[float, int, np.number, NDArray[np.number]],
        num_variables: int,
        mutation: str = "rand_1",
        F_min: float = 0.1,
        F_max: float = 0.9,
        t_F: float = 0.1,
        t_CR: float = 0.1,
        elitism: bool = True,
        init_population: Optional[NDArray[np.float64]] = None,
        genotype_to_phenotype: Optional[Callable[[NDArray[np.float64]], NDArray[Any]]] = None,
        optimal_value: Optional[float] = None,
        termination_error_value: float = 0.0,
        no_increase_num: Optional[int] = None,
        minimization: bool = False,
        show_progress_each: Optional[int] = None,
        keep_history: bool = False,
        n_jobs: int = 1,
        fitness_function_args: Optional[Dict] = None,
        genotype_to_phenotype_args: Optional[Dict] = None,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        on_generation: Optional[Callable] = None,
    ):
        DifferentialEvolution.__init__(
            self,
            fitness_function=fitness_function,
            iters=iters,
            pop_size=pop_size,
            left_border=left_border,
            right_border=right_border,
            num_variables=num_variables,
            mutation=mutation,
            elitism=elitism,
            init_population=init_population,
            genotype_to_phenotype=genotype_to_phenotype,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history,
            n_jobs=n_jobs,
            fitness_function_args=fitness_function_args,
            genotype_to_phenotype_args=genotype_to_phenotype_args,
            random_state=random_state,
            on_generation=on_generation,
        )

        self._F_min: float = F_min
        self._F_max: float = F_max
        self._t_F: float = t_F
        self._t_CR: float = t_CR

        self._F: NDArray[np.float64] = np.full(self._pop_size, 0.5, dtype=np.float64)
        self._CR: NDArray[np.float64] = np.full(self._pop_size, 0.9, dtype=np.float64)

    def _get_mutate_F(self: jDE) -> NDArray[np.float64]:
        mutate_F = self._F.copy()
        mask = uniform(0.0, 1.0, size=self._pop_size) < self._t_F

        random_values = uniform(0.0, 1.0, size=np.sum(mask))
        mutate_F[mask] = self._F_min + random_values * self._F_max
        return mutate_F

    def _get_mutate_CR(self: jDE) -> NDArray[np.float64]:
        mutate_CR = self._CR.copy()
        mask = uniform(0.0, 1.0, size=self._pop_size) < self._t_CR

        random_values = uniform(0.0, 1.0, size=np.sum(mask))
        mutate_CR[mask] = random_values
        return mutate_CR

    def _get_new_population(self: jDE) -> None:
        get_new_individ_g = partial(
            self._get_new_individ_g,
        )

        mutate_F = self._get_mutate_F()
        mutate_CR = self._get_mutate_CR()

        mutant_cr_b_g = np.array(
            [
                get_new_individ_g(individ_g=self._population_g_i[i], F=mutate_F[i], CR=mutate_CR[i])
                for i in range(self._pop_size)
            ],
            dtype=np.float64,
        )

        mutant_cr_ph = self._get_phenotype(mutant_cr_b_g)
        mutant_cr_fit = self._get_fitness(mutant_cr_ph)
        mask = mutant_cr_fit >= self._fitness_i

        self._population_g_i[mask] = mutant_cr_b_g[mask]
        self._population_ph_i[mask] = mutant_cr_ph[mask]
        self._fitness_i[mask] = mutant_cr_fit[mask]
        self._F[mask] = mutate_F[mask]
        self._CR[mask] = mutate_CR[mask]

    def _update_data(self: jDE) -> None:
        super()._update_data()
        self._update_stats(F=self._F, CR=self._CR)



from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
from numpy.typing import NDArray

from ._geneticprogramming import GeneticProgramming
from ._pdpga import PDPGA
from ..base import Tree
from ..base import UniversalSet


class PDPGP(GeneticProgramming, PDPGA):
    """Niehaus, J., Banzhaf, W. (2001). Adaption of Operator Probabilities in
    Genetic Programming. In: Miller, J., Tomassini, M., Lanzi, P.L., Ryan, C.,
    Tettamanzi, A.G.B., Langdon, W.B. (eds) Genetic Programming. EuroGP 2001.
    Lecture Notes in Computer Science, vol 2038. Springer, Berlin, Heidelberg.
    https://doi.org/10.1007/3-540-45355-5_26
    """

    def __init__(
        self,
        fitness_function: Callable[[NDArray[Any]], NDArray[np.float64]],
        uniset: UniversalSet,
        iters: int,
        pop_size: int,
        tour_size: int = 2,
        mutation_rate: float = 0.05,
        parents_num: int = 2,
        elitism: bool = True,
        selections: Tuple[str, ...] = (
            "proportional",
            "rank",
            "tournament_3",
            "tournament_5",
            "tournament_7",
        ),
        crossovers: Tuple[str, ...] = ("gp_standard", "gp_one_point", "gp_uniform_rank_2"),
        mutations: Tuple[str, ...] = (
            "gp_weak_point",
            "gp_average_point",
            "gp_strong_point",
            "gp_weak_grow",
            "gp_average_grow",
            "gp_strong_grow",
        ),
        max_level: int = 16,
        init_level: int = 4,
        init_population: Optional[NDArray] = None,
        genotype_to_phenotype: Optional[Callable[[NDArray], NDArray[Any]]] = None,
        optimal_value: Optional[float] = None,
        termination_error_value: float = 0.0,
        no_increase_num: Optional[int] = None,
        minimization: bool = False,
        show_progress_each: Optional[int] = None,
        keep_history: bool = False,
        n_jobs: int = 1,
        fitness_function_args: Optional[Dict] = None,
        genotype_to_phenotype_args: Optional[Dict] = None,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        on_generation: Optional[Callable] = None,
    ):
        PDPGA.__init__(
            self,
            fitness_function=fitness_function,
            iters=iters,
            pop_size=pop_size,
            str_len=1,
            tour_size=tour_size,
            mutation_rate=mutation_rate,
            parents_num=parents_num,
            elitism=elitism,
            selections=selections,
            crossovers=crossovers,
            mutations=mutations,
            init_population=init_population,
            genotype_to_phenotype=genotype_to_phenotype,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history,
            n_jobs=n_jobs,
            fitness_function_args=fitness_function_args,
            genotype_to_phenotype_args=genotype_to_phenotype_args,
            random_state=random_state,
            on_generation=on_generation,
        )

        GeneticProgramming.__init__(
            self,
            fitness_function=fitness_function,
            uniset=uniset,
            iters=iters,
            pop_size=pop_size,
            tour_size=tour_size,
            mutation_rate=mutation_rate,
            parents_num=parents_num,
            elitism=elitism,
            max_level=max_level,
            init_level=init_level,
            init_population=init_population,
            genotype_to_phenotype=genotype_to_phenotype,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history,
            n_jobs=n_jobs,
            fitness_function_args=fitness_function_args,
            genotype_to_phenotype_args=genotype_to_phenotype_args,
            random_state=random_state,
            on_generation=on_generation,
        )

    def _get_new_individ_g(
        self: PDPGP,
        specified_selection: str,
        specified_crossover: str,
        specified_mutation: str,
    ) -> Tree:
        selection_func, tour_size = self._selection_pool[specified_selection]
        crossover_func, quantity = self._crossover_pool[specified_crossover]
        mutation_func, proba, is_constant_rate = self._mutation_pool[specified_mutation]

        selected_id = selection_func(
            self._fitness_scale_i, self._fitness_rank_i, np.int64(tour_size), np.int64(quantity)
        )

        previous_fitness = self._choice_parent(self._fitness_i[selected_id])
        self._previous_fitness_i.append(previous_fitness)

        offspring_no_mutated = crossover_func(
            self._population_g_i[selected_id],
            self._fitness_scale_i[selected_id],
            self._fitness_rank_i[selected_id],
            self._max_level,
        )

        if is_constant_rate:
            proba = proba
        else:
            proba = proba / len(offspring_no_mutated)

        offspring = mutation_func(offspring_no_mutated, self._uniset, proba, self._max_level)
        return offspring



from __future__ import annotations

from functools import partial
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Union

from numba import float64
from numba import njit

import numpy as np
from numpy.typing import NDArray

from ..base._ea import EvolutionaryAlgorithm
from ..utils import if_single_or_array_to_float_array
from ..utils.mutations import best_1
from ..utils.mutations import best_2
from ..utils.crossovers import binomial
from ..utils.mutations import current_to_best_1
from ..utils.mutations import rand_1
from ..utils.mutations import rand_2
from ..utils.mutations import rand_to_best1
from ..utils.random import uniform


@njit(float64[:](float64[:], float64[:], float64[:]))
def bounds_control(
    array: NDArray[np.float64], left: NDArray[np.float64], right: NDArray[np.float64]
) -> NDArray[np.float64]:
    to_return = array.copy()

    size = len(array)
    for i in range(size):
        if array[i] < left[i]:
            to_return[i] = left[i]
        elif array[i] > right[i]:
            to_return[i] = right[i]
    return to_return


class DifferentialEvolution(EvolutionaryAlgorithm):
    """Storn, Rainer & Price, Kenneth. (1995). Differential Evolution: A Simple and Efficient
    Adaptive Scheme for Global Optimization Over Continuous Spaces.
    Journal of Global Optimization. 23"""

    def __init__(
        self,
        fitness_function: Callable[[NDArray[Any]], NDArray[np.float64]],
        iters: int,
        pop_size: int,
        left_border: Union[float, int, np.number, NDArray[np.number]],
        right_border: Union[float, int, np.number, NDArray[np.number]],
        num_variables: int,
        mutation: str = "rand_1",
        F: float = 0.5,
        CR: float = 0.5,
        elitism: bool = True,
        init_population: Optional[NDArray[np.float64]] = None,
        genotype_to_phenotype: Optional[Callable[[NDArray[np.float64]], NDArray[Any]]] = None,
        optimal_value: Optional[float] = None,
        termination_error_value: float = 0.0,
        no_increase_num: Optional[int] = None,
        minimization: bool = False,
        show_progress_each: Optional[int] = None,
        keep_history: bool = False,
        n_jobs: int = 1,
        fitness_function_args: Optional[Dict] = None,
        genotype_to_phenotype_args: Optional[Dict] = None,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        on_generation: Optional[Callable] = None,
    ):
        EvolutionaryAlgorithm.__init__(
            self,
            fitness_function=fitness_function,
            iters=iters,
            pop_size=pop_size,
            elitism=elitism,
            init_population=init_population,
            genotype_to_phenotype=genotype_to_phenotype,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history,
            n_jobs=n_jobs,
            fitness_function_args=fitness_function_args,
            genotype_to_phenotype_args=genotype_to_phenotype_args,
            random_state=random_state,
            on_generation=on_generation,
        )

        self._num_variables = num_variables
        self._left: NDArray[np.float64] = if_single_or_array_to_float_array(
            left_border, self._num_variables
        )
        self._right: NDArray[np.float64] = if_single_or_array_to_float_array(
            right_border, self._num_variables
        )
        self._specified_mutation: str = mutation
        self._F: Union[float, NDArray[np.float64]] = F
        self._CR: Union[float, NDArray[np.float64]] = CR

        self._mutation_pool: Dict[str, Callable] = {
            "best_1": best_1,
            "rand_1": rand_1,
            "current_to_best_1": current_to_best_1,
            "rand_to_best1": rand_to_best1,
            "best_2": best_2,
            "rand_2": rand_2,
        }

    @staticmethod
    def float_population(
        pop_size: int,
        left_border: Union[float, int, np.number, NDArray[np.number]],
        right_border: Union[float, int, np.number, NDArray[np.number]],
        num_variables: int,
    ) -> NDArray[np.float64]:
        left = if_single_or_array_to_float_array(left_border, num_variables)
        right = if_single_or_array_to_float_array(right_border, num_variables)
        points_along_axis = [
            uniform(left_i, right_i, pop_size) for left_i, right_i in zip(left, right)
        ]
        return np.array(points_along_axis, dtype=np.float64).T

    def _first_generation(self: DifferentialEvolution) -> None:
        if self._init_population is None:
            self._population_g_i = self.float_population(
                pop_size=self._pop_size,
                left_border=self._left,
                right_border=self._right,
                num_variables=self._num_variables,
            )
        else:
            self._population_g_i = self._init_population.copy()

    def _get_init_population(self: DifferentialEvolution) -> None:
        self._first_generation()
        self._population_ph_i = self._get_phenotype(self._population_g_i)
        self._fitness_i = self._get_fitness(self._population_ph_i)

    def _get_new_individ_g(
        self: DifferentialEvolution,
        individ_g: NDArray[np.float64],
        F: float,
        CR: float,
    ) -> NDArray[np.float64]:
        mutation_func = self._mutation_pool[self._specified_mutation]

        mutant_g = mutation_func(
            individ_g, self._thefittest._genotype, self._population_g_i, np.float64(F)
        )

        mutant_cr_g = binomial(individ_g, mutant_g, np.float64(CR))
        return bounds_control(mutant_cr_g, self._left, self._right)

    def _get_new_population(self: DifferentialEvolution) -> None:
        get_new_individ_g = partial(
            self._get_new_individ_g,
            F=self._F,
            CR=self._CR,
        )

        mutant_cr_b_g = np.array(
            [get_new_individ_g(individ_g=self._population_g_i[i]) for i in range(self._pop_size)],
            dtype=np.float64,
        )

        mutant_cr_ph = self._get_phenotype(mutant_cr_b_g)
        mutant_cr_fit = self._get_fitness(mutant_cr_ph)
        mask = mutant_cr_fit >= self._fitness_i

        self._population_g_i[mask] = mutant_cr_b_g[mask]
        self._population_ph_i[mask] = mutant_cr_ph[mask]
        self._fitness_i[mask] = mutant_cr_fit[mask]

    def _from_population_g_to_fitness(self: EvolutionaryAlgorithm) -> None:
        self._update_data()

        if self._elitism:
            (
                self._population_g_i[-1],
                self._population_ph_i[-1],
                self._fitness_i[-1],
            ) = self._thefittest.get().values()

        self._adapt()



import random
from typing import List
from typing import Tuple
from typing import Union
from typing import Optional
import numbers

from numba import boolean
from numba import float64
from numba import int64
from numba import njit

import numpy as np
from numpy.typing import NDArray

from ..utils import binary_search_interval
from ..utils import check_for_value


@njit(int64[:](int64[:]))
def sattolo_shuffle(arr):
    """
    Perform Sattolo's algorithm for in-place array shuffling.

    Parameters
    ----------
    arr : int64[:]
        Input array to be shuffled.

    Returns
    -------
    int64[:]
        Shuffled array.

    Examples
    --------
    >>> from numba import jit
    >>> import numpy as np
    >>> from thefittest.utils.random import sattolo_shuffle
    >>> # Example of using Sattolo's shuffle
    >>> arr = np.array([1, 2, 3, 4, 5], dtype = np.int64)
    >>> print("Shuffled Array:", sattolo_shuffle(arr))
    Shuffled Array: ...

    Notes
    -----
    Sattolo's algorithm generates a random cyclic permutation of a given array.
    """
    n = len(arr)
    shuffled_arr = arr.copy()

    for i in range(n - 1, 0, -1):
        j = np.int64(np.floor(random.random() * i))
        shuffled_arr[i], shuffled_arr[j] = shuffled_arr[j], shuffled_arr[i]

    return shuffled_arr


@njit
def sattolo_shuffle_2d(arr):
    """
    Perform Sattolo's algorithm for in-place shuffling of rows in a 2D array.

    Parameters
    ----------
    arr : int64[:, :]
        Input 2D array to be shuffled.

    Returns
    -------
    int64[:, :]
        Shuffled 2D array.

    Examples
    --------
    >>> from numba import jit
    >>> import numpy as np
    >>> from thefittest.utils.random import sattolo_shuffle_2d
    >>> # Example of using Sattolo's shuffle for a 2D array
    >>> arr_2d = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    >>> print("Shuffled 2D Array:", sattolo_shuffle_2d(arr_2d))
    Shuffled 2D Array: ...

    Notes
    -----
    Sattolo's algorithm generates a random cyclic permutation of rows in a 2D array.
    """
    n_rows, _ = arr.shape
    shuffled_arr = arr.copy()

    for i in range(n_rows - 1, 0, -1):
        j = np.int64(np.floor(random.random() * i))
        shuffled_arr[i], shuffled_arr[j] = shuffled_arr[j].copy(), shuffled_arr[i].copy()

    return shuffled_arr


@njit
def numba_seed(seed_value):
    random.seed(seed_value)
    np.random.seed(seed_value)


@njit(float64[:](float64, float64, int64))
def cauchy_distribution(loc: np.float64, scale: np.float64, size: np.int64) -> NDArray[np.float64]:
    """
    Generate an array of random numbers from a Cauchy distribution.

    Parameters
    ----------
    loc : np.float64
        The location parameter of the Cauchy distribution.
    scale : np.float64
        The scale parameter of the Cauchy distribution.
    size : np.int64
        The size of the array to generate.

    Returns
    -------
    NDArray[np.float64]
        An array of random numbers drawn from a Cauchy distribution.

    Examples
    --------
    >>> import numpy as np
    >>> from thefittest.utils.random import cauchy_distribution
    >>>
    >>> # Generate an array of 10 random numbers from a Cauchy distribution
    >>> loc_value = 0.0
    >>> scale_value = 1.0
    >>> size_value = 10
    >>> cauchy_result = cauchy_distribution(loc_value, scale_value, size_value)
    >>>
    >>> print("Cauchy Distribution Result:", cauchy_result)
    Cauchy Distribution Result: ...
    """

    x_ = np.random.standard_cauchy(size=size).astype(np.float64)
    return loc + scale * x_


@njit(int64[:](float64[:], int64, boolean))
def random_weighted_sample(
    weights: NDArray[np.float64], quantity: Union[np.int64, int] = 1, replace: bool = True
) -> NDArray[np.int64]:
    """
    Generate a random weighted sample.

    Parameters
    ----------
    weights : NDArray[np.float64]
        1D array of weights representing the probability of each element being selected.
    quantity : Union[np.int64, int]
        The number of elements to sample. Default is 1.
    replace : bool
        Whether sampling is done with replacement. Default is True.

    Returns
    -------
    NDArray[np.int64]
        An array of sampled indices.

    Examples
    --------
    >>> from thefittest.utils.random import random_weighted_sample
    >>> import numpy as np
    >>>
    >>> # Example with replacement
    >>> weights = np.array([0.3, 0.2, 0.5])
    >>> sampled_indices = random_weighted_sample(weights, quantity=2, replace=True)
    >>> print("Sampled Indices:", sampled_indices)
    Sampled Indices: ...
    >>>
    >>> # Example without replacement
    >>> sampled_indices_no_replace = random_weighted_sample(weights, quantity=2, replace=False)
    >>> print("Sampled Indices (No Replace):", sampled_indices_no_replace)
    Sampled Indices (No Replace): ...
    """

    if not replace:
        assert len(weights) >= quantity
    sample = np.empty(quantity, dtype=np.int64)

    cumsumweights = np.cumsum(weights)
    sumweights = cumsumweights[-1]

    i = 0
    while i < quantity:
        roll = sumweights * random.random()
        ind = binary_search_interval(roll, cumsumweights)
        if not replace:
            if check_for_value(ind, sample, i):
                continue

        sample[i] = ind
        i += 1
    return sample


@njit(int64[:](int64, int64, boolean))
def random_sample(
    range_size: np.int64, quantity: np.int64, replace: bool = True
) -> NDArray[np.int64]:
    """
    Generate a random sample from a range.

    Parameters
    ----------
    range_size : np.int64
        The size of the range to sample from.
    quantity : np.int64
        The number of elements to sample.
    replace : bool
        Whether sampling is done with replacement. Default is True.

    Returns
    -------
    NDArray[np.int64]
        An array of sampled indices.

    Examples
    --------
    >>> from thefittest.utils.random import random_sample
    >>>
    >>> # Example with replacement
    >>> sampled_indices = random_sample(range_size=10, quantity=3, replace=True)
    >>> print("Sampled Indices:", sampled_indices)
    Sampled Indices: ...
    >>>
    >>> # Example without replacement
    >>> sampled_indices_no_replace = random_sample(range_size=10, quantity=3, replace=False)
    >>> print("Sampled Indices (No Replace):", sampled_indices_no_replace)
    Sampled Indices (No Replace): ...
    """

    if not replace:
        assert range_size >= quantity
    sample = np.empty(quantity, dtype=np.int64)
    i = 0
    while i < quantity:
        # ind = random.randrange(int(range_size))
        ind = np.random.randint(0, int(range_size))

        if not replace:
            if check_for_value(ind, sample, i):
                continue

        sample[i] = ind
        i += 1
    return sample


@njit(boolean(float64))
def flip_coin(threshold):
    """
    Simulate a biased coin flip.

    Parameters
    ----------
    threshold : float64
        The threshold for the biased coin flip. Should be in the range [0, 1].

    Returns
    -------
    boolean
        Returns True with a probability equal to the given threshold and
        False with a complementary probability.

    Examples
    --------
    >>> from thefittest.utils.random import flip_coin
    >>>
    >>> # Example of a biased coin flip with a threshold of 0.5
    >>> result = flip_coin(0.5)
    >>> print("Coin Flip Result:", result)
    Coin Flip Result: ...

    Notes
    -----
    The function simulates a biased coin flip with a specified threshold.
    If the threshold is 0.5, it behaves like a fair coin flip.
    A threshold greater than 0.5 biases the result towards True,
    while a threshold less than 0.5 biases the result towards False.
    """
    return random.random() < threshold


@njit(float64[:](float64, float64, int64))
def uniform(low: np.float64, high: np.float64, size: np.int64):
    """
    Generate an array of random samples from a uniform distribution.

    Parameters
    ----------
    low : np.float64
        The lower boundary of the output interval. All values generated will be
        greater than or equal to `low`.
    high : np.float64
        The upper boundary of the output interval. All values generated will be
        less or equal to `high`.
    size : np.int64
        The number of elements to generate.

    Returns
    -------
    NDArray[np.float64]
        An 1D array of random samples drawn from the uniform distribution.

    Examples
    --------
    >>> from thefittest.utils.random import uniform
    >>>
    >>> # Example of generating random samples
    >>> result = uniform(low=0.0, high=1.0, size=5)
    >>> print("Random Samples:", result)
    Random Samples: ...

    Notes
    -----
    The generated samples follow a uniform distribution, where each value within
    the specified range has an equal probability of being selected.

    """
    return np.random.uniform(low=low, high=high, size=size)


@njit(int64[:](int64, int64, int64))
def randint(low, high, size):
    """
    Generate an array of random integers from a discrete uniform distribution.

    Parameters
    ----------
    low : int
        The lowest integer to be drawn from the distribution.
    high : int
        The highest integer to be drawn from the distribution.
    size : int
        The number of integers to generate.

    Returns
    -------
    NDArray[int64]
        An array of random integers.

    Examples
    --------
    >>> from numba import jit
    >>> import numpy as np
    >>> from thefittest.utils.random import randint
    >>> # Example of generating random integers
    >>> result = randint(low=1, high=10, size=5)
    >>> print("Random Integers:", result)
    Random Integers: ...

    Notes
    -----
    The generated integers follow a discrete uniform distribution.
    """
    result = np.empty(size, dtype=np.int64)

    for i in range(size):
        result[i] = low + np.int64(np.floor((high - low) * random.random()))

    return result


def check_random_state(seed: Optional[Union[int, np.random.RandomState]] = None):
    """Turn seed into a np.random.RandomState instance

    Parameters
    ----------
    seed : None | int | instance of RandomState
        If seed is None, return the RandomState singleton used by np.random.
        If seed is an int, return a new RandomState instance seeded with seed.
        If seed is already a RandomState instance, return it.
        Otherwise raise ValueError.

    """
    if seed is None:
        random_state = np.random.mtrand._rand
    elif isinstance(seed, (numbers.Integral, np.integer)):
        random_state = np.random.RandomState(seed)
        seed = random_state.get_state()[1][0]
        numba_seed(seed)
    elif isinstance(seed, np.random.RandomState):
        random_state = seed
        seed = random_state.get_state()[1][0]
        numba_seed(seed)
    else:
        raise ValueError("%r cannot be used to seed a numpy.random.RandomState" " instance" % seed)

    return random_state


def generator1() -> float:
    value = np.round(uniform(0, 10, 1)[0], 4)
    return value


def generator2() -> int:
    value = randint(0, 10, 1)[0]
    return value



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

import numpy as np
from numpy.typing import ArrayLike
from numpy.typing import NDArray

from sklearn.base import BaseEstimator
from sklearn.base import ClassifierMixin
from sklearn.preprocessing import LabelEncoder
from sklearn.preprocessing import OneHotEncoder
from sklearn.utils.multiclass import check_classification_targets
from sklearn.utils.validation import check_X_y

from ..base import Net
from ..base._ea import Statistics
from ..base._net import ACTIV_NAME_INV
from ..optimizers import DifferentialEvolution
from ..optimizers import GeneticAlgorithm
from ..optimizers import SHADE
from ..optimizers import SHAGA
from ..optimizers import SelfCGA
from ..optimizers import jDE
from ..utils import (
    array_like_to_numpy_X_y,
    _is_torch_optimizer,
)
from ..utils._metrics import _ce_from_probs_torch
from ..utils._metrics import _rmse_torch
from ..utils.random import check_random_state
from ..utils.transformations import GrayCode
import torch
import warnings
from torch.optim import Optimizer as TorchOptimizer


weights_type_optimizer_alias = Union[
    Type[DifferentialEvolution],
    Type[jDE],
    Type[SHADE],
    Type[GeneticAlgorithm],
    Type[SelfCGA],
    Type[SHAGA],
    Type[TorchOptimizer],
]
weights_optimizer_alias = Union[DifferentialEvolution, jDE, SHADE, GeneticAlgorithm, SelfCGA, SHAGA]


def check_optimizer_args(
    optimizer_args, args_auto_defined: Optional[list] = None, args_in_class: Optional[list] = None
) -> None:
    if args_auto_defined is not None:
        for arg in args_auto_defined:
            assert (
                arg not in optimizer_args.keys()
            ), f"""Do not set the "{arg}"
            to the "weights_optimizer_args". It is defined automatically"""

    if args_in_class is not None:
        for arg in args_in_class:
            assert (
                arg not in optimizer_args.keys()
            ), f"Do not set '{arg}' in 'optimizer_args'. Instead, use the arguments of the class."


def fitness_function_weights(
    weights: NDArray[np.float64],
    net: "Net",
    X: torch.Tensor,
    targets: torch.Tensor,
    task_type: str = "regression",
) -> NDArray[np.float64]:

    weights_t = torch.as_tensor(weights, dtype=X.dtype, device=X.device)
    pop_size = weights.shape[0]
    losses = np.empty(pop_size, dtype=np.float64)

    with torch.no_grad():
        if task_type == "classification":
            for i in range(pop_size):
                out = net.forward(X, weights_t[i])
                loss = _ce_from_probs_torch(out, targets)
                losses[i] = float(loss.item())
        elif task_type == "regression":
            for i in range(pop_size):
                out = net.forward(X, weights_t[i])
                loss = _rmse_torch(out, targets)
                losses[i] = float(loss.item())
        else:
            raise ValueError("task_type must be 'classification' or 'regression'")
    return losses


def train_net_weights(
    net: Net,
    X_train: NDArray[np.float64],
    y_train: NDArray[np.float64],
    weights_optimizer_args: Dict,
    weights_optimizer: weights_type_optimizer_alias,
    fitness_function: Callable,
    task_type: str = "regression",
) -> Net:
    net = net.copy()
    net.compile_torch(X_train.device)

    if _is_torch_optimizer(weights_optimizer):
        w_param = torch.nn.Parameter(
            net._weights.detach().to(dtype=X_train.dtype, device=X_train.device)
        )

        opt_args = dict(weights_optimizer_args) if weights_optimizer_args is not None else {}
        epochs = int(opt_args.pop("epochs", opt_args.pop("iters", 100)))
        show_each = opt_args.pop("show_progress_each", None)

        if show_each is not None:
            show_each = int(show_each)

        if "pop_size" in opt_args:
            warnings.warn(
                "`pop_size` игнорируется для torch.optim.* оптимизаторов.",
                RuntimeWarning,
            )
            opt_args.pop("pop_size", None)

        torch_opt = weights_optimizer([w_param], **opt_args)
        losses_history: list[float] = []

        for epoch in range(epochs):
            torch_opt.zero_grad(set_to_none=True)
            out = net.forward(X_train, w_param)
            if task_type == "classification":
                loss = _ce_from_probs_torch(out, y_train)
            elif task_type == "regression":
                loss = _rmse_torch(out, y_train)
            else:
                raise ValueError("task_type must be 'classification' or 'regression'")
            loss.backward()
            torch_opt.step()
            losses_history.append(float(loss.item()))

            if show_each is not None and (epoch % show_each == 0):
                print(f"[epoch {epoch}/{epochs}] loss={loss.item():.6f}", flush=True)

        weights = w_param.detach()
        return weights, None
    else:
        weights_optimizer_args = weights_optimizer_args.copy()

        weights_optimizer_args["fitness_function"] = fitness_function
        weights_optimizer_args["fitness_function_args"] = {
            "net": net,
            "X": X_train,
            "targets": y_train,
            "task_type": task_type,
        }

        initial_population: Union[NDArray[np.float64], NDArray[np.byte]] = (
            DifferentialEvolution.float_population(
                weights_optimizer_args["pop_size"], -10, 10, len(net._weights)
            )
        )
        initial_population[0] = net._weights.cpu().numpy().copy()

        if weights_optimizer in (SHADE, DifferentialEvolution, jDE):
            weights_optimizer_args["left_border"] = -10
            weights_optimizer_args["right_border"] = 10
            weights_optimizer_args["num_variables"] = len(net._weights)
        else:
            genotype_to_phenotype = GrayCode().fit(
                left_border=-10.0,
                right_border=10.0,
                num_variables=len(net._weights),
                bits_per_variable=16,
            )
            weights_optimizer_args["str_len"] = np.sum(genotype_to_phenotype._bits_per_variable)
            weights_optimizer_args["genotype_to_phenotype"] = genotype_to_phenotype.transform

        weights_optimizer_args["minimization"] = True
        optimizer = weights_optimizer(**weights_optimizer_args)
        optimizer.fit()

        weights = torch.as_tensor(
            optimizer.get_fittest()["phenotype"], dtype=X_train.dtype, device=X_train.device
        )

        return weights, optimizer._stats


class BaseMLPEA(BaseEstimator, metaclass=ABCMeta):
    """
    Attributes that have been estimated from the data must always have a name ending with trailing underscore,
    for example the coefficients of some regression estimator would be stored in a coef_ attribute after fit has been called.
    The estimated attributes are expected to be overridden when you call fit a second time.

    In iterative algorithms, the number of iterations should be specified by an integer called n_iter.
    """

    @abstractmethod
    def __init__(
        self,
        *,
        n_iter: int,
        pop_size: int,
        hidden_layers: Tuple[int, ...],
        activation: str = "sigma",
        offset: bool = True,
        weights_optimizer: weights_type_optimizer_alias = SHAGA,
        weights_optimizer_args: Optional[dict[str, Any]] = None,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        device: str = "cpu",
    ):
        super().__init__()
        self.n_iter = n_iter
        self.pop_size = pop_size
        self.hidden_layers = hidden_layers
        self.activation = activation
        self.offset = offset
        self.weights_optimizer = weights_optimizer
        self.weights_optimizer_args = weights_optimizer_args
        self.random_state = random_state
        self.device = device

    def _defitne_net(self, n_inputs: int, n_outputs: int) -> Net:
        start = 0
        end = n_inputs
        inputs_id = set(range(start, end))

        net = Net(inputs=inputs_id)

        for n_layer in self.hidden_layers:
            start = end
            end = end + n_layer
            inputs_id = {n_inputs - 1}
            hidden_id = set(range(start, end))
            activs = dict(zip(hidden_id, [ACTIV_NAME_INV[self.activation]] * len(hidden_id)))

            if self.offset:
                layer_net = Net(inputs=inputs_id) > Net(hidden_layers=[hidden_id], activs=activs)
            else:
                layer_net = Net(hidden_layers=[hidden_id], activs=activs)

            net = net > layer_net

        start = end
        end = end + n_outputs
        inputs_id = {n_inputs - 1}
        output_id = set(range(start, end))

        if isinstance(self, ClassifierMixin):
            outputs_activation = [ACTIV_NAME_INV["softmax"]] * len(output_id)
        else:
            outputs_activation = [ACTIV_NAME_INV["ln"]] * len(output_id)

        activs = dict(zip(output_id, outputs_activation))

        if self.offset:
            layer_net = Net(inputs=inputs_id) > Net(outputs=output_id, activs=activs)
        else:
            layer_net = Net(outputs=output_id, activs=activs)

        net = net > layer_net
        net._offset = self.offset
        return net

    def get_stats(self) -> Statistics:
        return self.optimizer_stats_

    def get_net(self) -> Net:
        return self.net_

    def fit(self, X: ArrayLike, y: ArrayLike):
        if self.weights_optimizer_args is not None:
            weights_optimizer_args = self.weights_optimizer_args.copy()
            check_optimizer_args(
                weights_optimizer_args,
                args_auto_defined=[
                    "fitness_function",
                    "fitness_function_args",
                    "left_border",
                    "right_border",
                    "num_variables",
                    "str_len",
                    "genotype_to_phenotype",
                    "genotype_to_phenotype_args",
                    "minimization",
                    "init_population",
                ],
                args_in_class=["iters", "pop_size"],
            )
        else:
            weights_optimizer_args = {}

        weights_optimizer_args["iters"] = self.n_iter
        weights_optimizer_args["pop_size"] = self.pop_size

        check_random_state(self.random_state)

        if isinstance(self, ClassifierMixin):
            X, y = check_X_y(X, y)
            check_classification_targets(y)

            self._label_encoder = LabelEncoder()
            self._one_hot_encoder = OneHotEncoder(
                sparse_output=False, categories="auto", dtype=np.float64
            )

            numeric_labels = self._label_encoder.fit_transform(y)
            y = self._one_hot_encoder.fit_transform(np.array(numeric_labels).reshape(-1, 1))
            self.classes_ = self._label_encoder.classes_
            self.n_classes_ = len(self.classes_)
        else:
            X, y = check_X_y(X, y)

        X, y = array_like_to_numpy_X_y(X, y)

        self.n_features_in_ = X.shape[1]

        if self.offset:
            X = np.hstack([X, np.ones((X.shape[0], 1))])

        device = torch.device(self.device)
        X_t = torch.as_tensor(X, dtype=torch.float32, device=device)
        y_t = torch.as_tensor(y, dtype=torch.float32, device=device)

        if isinstance(self, ClassifierMixin):
            self.net_ = self._defitne_net(X_t.shape[1], len(self.classes_))
            self.net_._weights, self.optimizer_stats_ = train_net_weights(
                self.net_,
                X_t,
                y_t,
                weights_optimizer_args,
                self.weights_optimizer,
                fitness_function_weights,
                task_type="classification",
            )
        else:
            self.net_ = self._defitne_net(X_t.shape[1], 1)
            self.net_._weights, self.optimizer_stats_ = train_net_weights(
                self.net_,
                X_t,
                y_t,
                weights_optimizer_args,
                self.weights_optimizer,
                fitness_function_weights,
                task_type="regression",
            )

        self.net_.compile_torch()
        return self



from __future__ import annotations

from typing import Any
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

import numpy as np
from numpy.typing import NDArray

from sklearn.base import ClassifierMixin
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.validation import validate_data

from ..base._gp import BaseGP
from ..base._gp import safe_sigmoid
from ..optimizers import GeneticProgramming
from ..optimizers import SelfCGP


class GeneticProgrammingClassifier(ClassifierMixin, BaseGP):
    def __init__(
        self,
        *,
        n_iter: int = 300,
        pop_size: int = 1000,
        functional_set_names: Tuple[str, ...] = ("cos", "sin", "add", "sub", "mul", "div"),
        optimizer: Union[Type[SelfCGP], Type[GeneticProgramming]] = SelfCGP,
        optimizer_args: Optional[dict[str, Any]] = None,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        use_fitness_cache: bool = False,
    ):
        super().__init__(
            n_iter=n_iter,
            pop_size=pop_size,
            functional_set_names=functional_set_names,
            optimizer=optimizer,
            optimizer_args=optimizer_args,
            random_state=random_state,
            use_fitness_cache=use_fitness_cache,
        )

    def predict_proba(self, X: NDArray[np.float64]):
        check_is_fitted(self)

        if hasattr(self, "_validate_data"):
            X = self._validate_data(X, reset=False)
        else:
            X = validate_data(self, X, reset=False)

        if X.shape[1] != self.n_features_in_:
            raise ValueError(
                f"X has {X.shape[1]} features, but this model was fitted with {self.n_features_in_}."
            )

        def _tree_out(tree):
            tfp = tree.set_terminals(**{f"x{i}": X[:, i] for i in range(self.n_features_in_)})
            return tfp() * np.ones(len(X))

        # OVR: если в fit обучили по дереву на класс
        if (
            hasattr(self, "trees_")
            and isinstance(self.trees_, (list, tuple))
            and len(self.trees_) > 0
        ):
            logits = np.column_stack([_tree_out(t) for t in self.trees_])  # (N, K)
            P = safe_sigmoid(logits)
            row_sums = P.sum(axis=1, keepdims=True)
            row_sums[row_sums == 0.0] = 1.0
            proba = (P / row_sums).astype(np.float64, copy=False)
            return proba

        tree_output = _tree_out(self.tree_)
        p1 = safe_sigmoid(tree_output)
        proba = np.vstack([1.0 - p1, p1]).T.astype(np.float64, copy=False)
        return proba

    def predict(self, X: NDArray[np.float64]):
        proba = self.predict_proba(X)  # ndarray (N, K)
        indeces = np.argmax(proba, axis=1)
        return self._label_encoder.inverse_transform(indeces)



from __future__ import annotations

from functools import partial
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

from numba import float64
from numba import njit

import numpy as np
from numpy.typing import NDArray

from ._differentialevolution import DifferentialEvolution
from ..utils import find_pbest_id
from ..utils.crossovers import binomial
from ..utils.mutations import current_to_pbest_1_archive_p_min
from ..utils.random import cauchy_distribution
from ..utils.random import randint
from ..utils.random import sattolo_shuffle_2d


@njit(float64[:](float64[:], float64[:], float64[:]))
def bounds_control_mean(
    array: NDArray[np.float64], left: NDArray[np.float64], right: NDArray[np.float64]
) -> NDArray[np.float64]:
    to_return = array.copy()
    size = len(array)
    for i in range(size):
        if array[i] < left[i]:
            to_return[i] = (left[i] + array[i]) / 2
        elif array[i] > right[i]:
            to_return[i] = (right[i] + array[i]) / 2
    return to_return


def lehmer_mean(
    x: NDArray[np.float64], power: int = 2, weight: Optional[NDArray[np.float64]] = None
) -> float:
    weight_arg: Union[NDArray[np.float64], int]
    if weight is None:
        weight_arg = 1
    else:
        weight_arg = weight

    x_up = weight_arg * np.power(x, power)
    x_down = weight_arg * np.power(x, power - 1)
    return np.sum(x_up) / np.sum(x_down)


@njit(float64(float64))
def randc01(u: np.float64) -> np.float64:
    """
    Generate a random number from a Cauchy distribution within the range (0, 1).

    Parameters
    ----------
    u : np.float64
        The location parameter for the Cauchy distribution.

    Returns
    -------
    np.float64
        A random number from a Cauchy distribution within the range (0, 1).

    Examples
    --------
    >>> import numpy as np
    >>> from thefittest.utils.random import randc01
    >>>
    >>> # Generate a random number from Cauchy distribution within the range (0, 1)
    >>> u_value = 0.0
    >>> result = randc01(u_value)
    >>>
    >>> print("Random Number from Cauchy Distribution (0, 1):", result)
    """

    value = cauchy_distribution(loc=u, scale=np.float64(0.1), size=np.int64(1))[0]
    while value <= 0:
        value = cauchy_distribution(loc=u, scale=np.float64(0.1), size=np.int64(1))[0]
    if value > 1:
        value = 1
    return value


@njit(float64(float64))
def randn01(u: np.float64) -> Union[float, np.float64]:
    """
    Generate a random number from a normal distribution within the range (0, 1).

    Parameters
    ----------
    u : np.float64
        The mean parameter for the normal distribution.

    Returns
    -------
    Union[float, np.float64]
        A random number from a normal distribution within the range (0, 1).

    Examples
    --------
    >>> import numpy as np
    >>> from thefittest.utils.random import randn01
    >>>
    >>> # Generate a random number from normal distribution within the range (0, 1)
    >>> u_value = 0.0
    >>> result = randn01(u_value)
    >>>
    >>> print("Random Number from Normal Distribution (0, 1):", result)
    """

    value = np.random.normal(u, 0.1, size=1)[0]
    if value < 0:
        return 0.0
    elif value > 1:
        return 1.0
    return value


class SHADE(DifferentialEvolution):
    """Tanabe, Ryoji & Fukunaga, Alex. (2013). Success-history based parameter adaptation
    for Differential Evolution. 2013 IEEE Congress on Evolutionary Computation,
    CEC 2013. 71-78. 10.1109/CEC.2013.6557555."""

    def __init__(
        self,
        fitness_function: Callable[[NDArray[Any]], NDArray[np.float64]],
        iters: int,
        pop_size: int,
        left_border: Union[float, int, np.number, NDArray[np.number]],
        right_border: Union[float, int, np.number, NDArray[np.number]],
        num_variables: int,
        elitism: bool = True,
        init_population: Optional[NDArray[np.float64]] = None,
        genotype_to_phenotype: Optional[Callable[[NDArray[np.float64]], NDArray[Any]]] = None,
        optimal_value: Optional[float] = None,
        termination_error_value: float = 0.0,
        no_increase_num: Optional[int] = None,
        minimization: bool = False,
        show_progress_each: Optional[int] = None,
        keep_history: bool = False,
        n_jobs: int = 1,
        fitness_function_args: Optional[Dict] = None,
        genotype_to_phenotype_args: Optional[Dict] = None,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        on_generation: Optional[Callable] = None,
    ):
        DifferentialEvolution.__init__(
            self,
            fitness_function=fitness_function,
            iters=iters,
            pop_size=pop_size,
            left_border=left_border,
            right_border=right_border,
            num_variables=num_variables,
            elitism=elitism,
            init_population=init_population,
            genotype_to_phenotype=genotype_to_phenotype,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history,
            n_jobs=n_jobs,
            fitness_function_args=fitness_function_args,
            genotype_to_phenotype_args=genotype_to_phenotype_args,
            random_state=random_state,
            on_generation=on_generation,
        )

        self._F: NDArray[np.float64]
        self._CR: NDArray[np.float64]
        self._population_g_archive_i: NDArray[np.float64] = np.zeros(shape=(0, len(self._left)))
        self._population_archive: NDArray[np.float64]
        self._pbest_id: NDArray[np.int64]
        self._H_size: int = pop_size
        self._H_F = np.full(self._H_size, 0.5, dtype=np.float64)
        self._H_CR = np.full(self._H_size, 0.5, dtype=np.float64)
        self._k: int = 0
        self._p: float = 0.05

    def _generate_F_CR(self: SHADE) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
        F_i = np.zeros(self._pop_size, dtype=np.float64)
        CR_i = np.zeros(self._pop_size, dtype=np.float64)
        for i in range(self._pop_size):
            r_i = randint(0, self._H_size, size=1)[0]
            u_F = self._H_F[r_i]
            u_CR = self._H_CR[r_i]
            F_i[i] = randc01(np.float64(u_F))
            CR_i[i] = randn01(np.float64(u_CR))
        return (F_i, CR_i)

    def _append_archive(
        self,
        archive: Union[NDArray[Any], NDArray[np.byte], NDArray[np.float64]],
        worse_g: Union[NDArray[Any], NDArray[np.byte], NDArray[np.float64]],
    ) -> Union[NDArray[Any], NDArray[np.byte], NDArray[np.float64]]:
        archive = np.append(archive, worse_g, axis=0)
        if len(archive) > self._pop_size:
            archive = sattolo_shuffle_2d(archive)

            archive = archive[: self._pop_size]
        return archive

    def _update_u_F(self, u_F: float, S_F: NDArray[np.float64]) -> float:
        if len(S_F):
            return lehmer_mean(S_F)
        return u_F

    def _update_u_CR(
        self, u_CR: float, S_CR: NDArray[np.float64], df: NDArray[np.float64]
    ) -> float:
        if len(S_CR):
            sum_ = np.sum(df)
            if sum_ > 0:
                weight_i = df / sum_
                return np.sum(weight_i * S_CR)
        return u_CR

    def _get_new_individ_g(
        self: SHADE,
        individ_g: NDArray[np.float64],
        F: float,
        CR: float,
    ) -> NDArray[np.float64]:
        mutant_g = current_to_pbest_1_archive_p_min(
            individ_g, self._population_g_i, self._pbest_id, F, self._population_archive
        )

        mutant_cr_g = binomial(individ_g, mutant_g, np.float64(CR))
        return bounds_control_mean(mutant_cr_g, self._left, self._right)

    def _get_new_population(self: SHADE) -> None:
        get_new_individ_g = partial(
            self._get_new_individ_g,
        )

        self._F, self._CR = self._generate_F_CR()
        self._pbest_id = find_pbest_id(self._fitness_i, np.float64(self._p))
        self._population_archive = np.vstack([self._population_g_i, self._population_g_archive_i])

        mutant_cr_b_g = np.array(
            [
                get_new_individ_g(individ_g=self._population_g_i[i], F=self._F[i], CR=self._CR[i])
                for i in range(self._pop_size)
            ],
            dtype=np.float64,
        )

        mutant_cr_ph = self._get_phenotype(mutant_cr_b_g)
        mutant_cr_fit = self._get_fitness(mutant_cr_ph)
        mask = mutant_cr_fit >= self._fitness_i
        succeses = mutant_cr_fit > self._fitness_i

        succeses_F = self._F[succeses]
        succeses_CR = self._CR[succeses]

        will_be_replaced_pop = self._population_g_i[succeses].copy()
        will_be_replaced_fit = self._fitness_i[succeses].copy()

        self._population_g_archive_i = self._append_archive(
            self._population_g_archive_i, will_be_replaced_pop
        )

        self._population_g_i[mask] = mutant_cr_b_g[mask]
        self._population_ph_i[mask] = mutant_cr_ph[mask]
        self._fitness_i[mask] = mutant_cr_fit[mask]

        d_fitness = np.abs(will_be_replaced_fit - self._fitness_i[succeses])

        if self._k + 1 == self._H_size:
            next_k = 0
        else:
            next_k = self._k + 1

        self._H_F[next_k] = self._update_u_F(self._H_F[self._k], succeses_F)
        self._H_CR[next_k] = self._update_u_CR(self._H_CR[self._k], succeses_CR, d_fitness)

        if self._k == self._H_size - 1:
            self._k = 0
        else:
            self._k += 1

    def _update_data(self: SHADE) -> None:
        super()._update_data()
        self._update_stats(H_F=self._H_F, H_CR=self._H_CR)



from __future__ import annotations

from functools import partial
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
from numpy.typing import NDArray

from scipy.stats import rankdata

from ..base import Tree
from ..base._ea import EvolutionaryAlgorithm
from ..utils.crossovers import empty_crossover
from ..utils.crossovers import empty_crossoverGP
from ..utils.mutations import flip_mutation
from ..utils.mutations import growing_mutation
from ..utils.crossovers import one_point_crossover
from ..utils.crossovers import one_point_crossoverGP
from ..utils.mutations import point_mutation
from ..utils.selections import proportional_selection
from ..utils.selections import rank_selection
from ..utils.mutations import shrink_mutation
from ..utils.crossovers import standard_crossover
from ..utils.mutations import swap_mutation
from ..utils.selections import tournament_selection
from ..utils.crossovers import two_point_crossover
from ..utils.crossovers import uniform_crossover
from ..utils.crossovers import uniform_crossoverGP
from ..utils.crossovers import uniform_proportional_crossover
from ..utils.crossovers import uniform_proportional_crossover_GP
from ..utils.crossovers import uniform_rank_crossover
from ..utils.crossovers import uniform_rank_crossover_GP
from ..utils.crossovers import uniform_tournament_crossover
from ..utils.crossovers import uniform_tournament_crossover_GP
from ..utils.transformations import minmax_scale
from ..utils.random import randint


class GeneticAlgorithm(EvolutionaryAlgorithm):
    """Holland, J. H. (1992). Genetic algorithms. Scientific American, 267(1), 66-72"""

    def __init__(
        self,
        fitness_function: Callable[[NDArray[Any]], NDArray[np.float64]],
        iters: int,
        pop_size: int,
        str_len: int,
        tour_size: int = 2,
        mutation_rate: float = 0.05,
        parents_num: int = 2,
        elitism: bool = True,
        selection: str = "tournament_5",
        crossover: str = "uniform_2",
        mutation: str = "weak",
        init_population: Optional[NDArray[np.byte]] = None,
        genotype_to_phenotype: Optional[Callable[[NDArray[np.byte]], NDArray[Any]]] = None,
        optimal_value: Optional[float] = None,
        termination_error_value: float = 0.0,
        no_increase_num: Optional[int] = None,
        minimization: bool = False,
        show_progress_each: Optional[int] = None,
        keep_history: bool = False,
        n_jobs: int = 1,
        fitness_function_args: Optional[Dict] = None,
        genotype_to_phenotype_args: Optional[Dict] = None,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        on_generation: Optional[Callable] = None,
    ):
        EvolutionaryAlgorithm.__init__(
            self,
            fitness_function=fitness_function,
            genotype_to_phenotype=genotype_to_phenotype,
            iters=iters,
            pop_size=pop_size,
            elitism=elitism,
            init_population=init_population,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history,
            n_jobs=n_jobs,
            fitness_function_args=fitness_function_args,
            genotype_to_phenotype_args=genotype_to_phenotype_args,
            random_state=random_state,
            on_generation=on_generation,
        )
        self._str_len: int = str_len
        self._tour_size: int = tour_size
        self._parents_num: int = parents_num
        self._mutation_rate: float = mutation_rate
        self._specified_selection: str = selection
        self._specified_crossover: str = crossover
        self._specified_mutation: str = mutation

        self._selection_pool: Dict[str, Tuple[Callable, Union[float, int]]] = {
            "proportional": (proportional_selection, 0),
            "rank": (rank_selection, 0),
            "tournament_k": (tournament_selection, self._tour_size),
            "tournament_3": (tournament_selection, 3),
            "tournament_5": (tournament_selection, 5),
            "tournament_7": (tournament_selection, 7),
        }

        self._crossover_pool: Dict[str, Tuple[Callable, Union[float, int]]] = {
            "empty": (empty_crossover, 1),
            "one_point": (one_point_crossover, 2),
            "two_point": (two_point_crossover, 2),
            "uniform_2": (uniform_crossover, 2),
            "uniform_7": (uniform_crossover, 7),
            "uniform_k": (uniform_crossover, self._parents_num),
            "uniform_prop_2": (uniform_proportional_crossover, 2),
            "uniform_prop_7": (uniform_proportional_crossover, 7),
            "uniform_prop_k": (uniform_proportional_crossover, self._parents_num),
            "uniform_rank_2": (uniform_rank_crossover, 2),
            "uniform_rank_7": (uniform_rank_crossover, 7),
            "uniform_rank_k": (uniform_rank_crossover, self._parents_num),
            "uniform_tour_3": (uniform_tournament_crossover, 3),
            "uniform_tour_7": (uniform_tournament_crossover, 7),
            "uniform_tour_k": (uniform_tournament_crossover, self._parents_num),
            "gp_empty": (empty_crossoverGP, 1),
            "gp_standard": (standard_crossover, 2),
            "gp_one_point": (one_point_crossoverGP, 2),
            "gp_uniform_2": (uniform_crossoverGP, 2),
            "gp_uniform_7": (uniform_crossoverGP, 7),
            "gp_uniform_k": (uniform_crossoverGP, self._parents_num),
            "gp_uniform_prop_2": (uniform_proportional_crossover_GP, 2),
            "gp_uniform_prop_7": (uniform_proportional_crossover_GP, 7),
            "gp_uniform_prop_k": (uniform_proportional_crossover_GP, self._parents_num),
            "gp_uniform_rank_2": (uniform_rank_crossover_GP, 2),
            "gp_uniform_rank_7": (uniform_rank_crossover_GP, 7),
            "gp_uniform_rank_k": (uniform_rank_crossover_GP, self._parents_num),
            "gp_uniform_tour_3": (uniform_tournament_crossover_GP, 3),
            "gp_uniform_tour_7": (uniform_tournament_crossover_GP, 7),
            "gp_uniform_tour_k": (uniform_tournament_crossover_GP, self._parents_num),
        }

        self._mutation_pool: Dict[str, Tuple[Callable, Union[float, int], bool]] = {
            "weak": (flip_mutation, 1 / 3, False),
            "average": (flip_mutation, 1, False),
            "strong": (flip_mutation, 3, False),
            "custom_rate": (flip_mutation, self._mutation_rate, True),
            "gp_weak_point": (point_mutation, 0.25, False),
            "gp_average_point": (point_mutation, 1, False),
            "gp_strong_point": (point_mutation, 4, False),
            "gp_custom_rate_point": (point_mutation, self._mutation_rate, True),
            "gp_weak_grow": (growing_mutation, 0.25, False),
            "gp_average_grow": (growing_mutation, 1, False),
            "gp_strong_grow": (growing_mutation, 4, False),
            "gp_custom_rate_grow": (growing_mutation, self._mutation_rate, True),
            "gp_weak_swap": (swap_mutation, 0.25, False),
            "gp_average_swap": (swap_mutation, 1, False),
            "gp_strong_swap": (swap_mutation, 4, False),
            "gp_custom_rate_swap": (swap_mutation, self._mutation_rate, True),
            "gp_weak_shrink": (shrink_mutation, 0.25, False),
            "gp_average_shrink": (shrink_mutation, 1, False),
            "gp_strong_shrink": (shrink_mutation, 4, False),
            "gp_custom_rate_shrink": (shrink_mutation, self._mutation_rate, True),
        }

        self._fitness_scale_i: NDArray[np.float64]
        self._fitness_rank_i: NDArray[np.float64]

    @staticmethod
    def binary_string_population(pop_size: int, str_len: int) -> NDArray[np.byte]:

        population = np.array(
            [randint(low=0, high=2, size=str_len) for _ in range(pop_size)],
            dtype=np.byte,
        )

        return population

    def _first_generation(self: GeneticAlgorithm) -> None:
        if self._init_population is None:
            self._population_g_i = self.binary_string_population(self._pop_size, self._str_len)
        else:
            self._population_g_i = self._init_population.copy()

    def _get_new_individ_g(
        self: GeneticAlgorithm,
        specified_selection: str,
        specified_crossover: str,
        specified_mutation: str,
    ) -> Union[Tree, np.byte]:
        selection_func, tour_size = self._selection_pool[specified_selection]
        crossover_func, quantity = self._crossover_pool[specified_crossover]
        mutation_func, proba, is_constant_rate = self._mutation_pool[specified_mutation]

        selected_id = selection_func(
            self._fitness_scale_i, self._fitness_rank_i, np.int64(tour_size), np.int64(quantity)
        )

        offspring_no_mutated = crossover_func(
            self._population_g_i[selected_id],
            self._fitness_scale_i[selected_id],
            self._fitness_rank_i[selected_id],
        )

        if is_constant_rate:
            proba = proba
        else:
            proba = proba / len(offspring_no_mutated)

        offspring = mutation_func(offspring_no_mutated, np.float64(proba))
        return offspring

    def _get_new_population(self: GeneticAlgorithm) -> None:
        get_new_individ_g = partial(
            self._get_new_individ_g,
            self._specified_selection,
            self._specified_crossover,
            self._specified_mutation,
        )

        self._population_g_i = np.array(
            [get_new_individ_g() for _ in range(self._pop_size)], dtype=self._population_g_i.dtype
        )

    def _from_population_g_to_fitness(self: GeneticAlgorithm) -> None:
        super()._from_population_g_to_fitness()

        self._fitness_scale_i = minmax_scale(self._fitness_i)
        self._fitness_rank_i = rankdata(self._fitness_i)

        self._adapt()



from __future__ import annotations

from typing import Any
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
from numpy.typing import ArrayLike

from sklearn.base import RegressorMixin
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.validation import validate_data

import torch

from ..base._mlp import BaseMLPEA
from ..base._mlp import weights_type_optimizer_alias
from ..optimizers import SHADE



class MLPEARegressor(RegressorMixin, BaseMLPEA):
    def __init__(
        self,
        *,
        n_iter: int = 200,
        pop_size: int = 500,
        hidden_layers: Tuple[int, ...] = (100,),
        activation: str = "sigma",
        offset: bool = True,
        weights_optimizer: weights_type_optimizer_alias = SHADE,
        weights_optimizer_args: Optional[dict[str, Any]] = None,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        device: str = "cpu",
    ):
        super().__init__(
            n_iter=n_iter,
            pop_size=pop_size,
            hidden_layers=hidden_layers,
            activation=activation,
            offset=offset,
            weights_optimizer=weights_optimizer,
            weights_optimizer_args=weights_optimizer_args,
            random_state=random_state,
            device=device,
        )

    def predict(self, X: ArrayLike):
        check_is_fitted(self)

        if hasattr(self, "_validate_data"):
            X = self._validate_data(X, reset=False)
        else:
            X = validate_data(self, X, reset=False)
        if self.offset:
            X = np.hstack([X, np.ones((X.shape[0], 1))])

        device = torch.device(self.device)
        X_t = torch.as_tensor(X, dtype=torch.float32, device=device)

        with torch.no_grad():
            out = self.net_.forward(X_t)

        if isinstance(out, torch.Tensor):
            out = out.detach().cpu().numpy().astype(np.float64)
        if out.ndim == 3 and out.shape[-1] == 1:
            out = out.squeeze(-1)
        if out.ndim == 2 and out.shape[-1] == 1:
            out = out.squeeze(-1)
        return out.reshape(-1)
